import logging
import sys
import threading
import time
import types

from pykit import threadutil

if sys.version_info[0] == 2:
    import Queue
else:
    import queue as Queue

logger = logging.getLogger(__name__)


class EmptyRst(object):
    pass


class Finish(object):
    pass


class JobWorkerError(Exception):
    pass


class JobWorkerNotFound(JobWorkerError):
    pass


class WorkerGroup(object):

    def __init__(self, index, worker, n_thread,
                 input_queue,
                 dispatcher,
                 probe, keep_order):

        self.index = index
        self.worker = worker
        self.n_thread = n_thread
        self.input_queue = input_queue
        self.output_queue = _make_q()
        self.dispatcher = dispatcher
        self.probe = probe
        self.keep_order = keep_order

        self.threads = {}

        # When exiting, it is not allowed to change thread number.
        # Because we need to send a `Finish` to each thread.
        self.exiting = False

        self.running = True

        # left close, right open: 0 running, n not.
        # Indicate what thread should have been running.
        self.running_index_range = [0, self.n_thread]

        # protect reading/writing worker_group info
        self.worker_group_lock = threading.RLock()

        # dispatcher mode implies keep_order
        if self.dispatcher is not None or self.keep_order:
            need_coordinator = True
        else:
            need_coordinator = False

        # Coordinator guarantees worker-group level first-in first-out, by put
        # output-queue into a queue-of-output-queue.
        if need_coordinator:
            # to maximize concurrency
            self.queue_of_output_q = _make_q(n=1024 * 1024)

            # Protect input.get() and ouput.put(), only used by non-dispatcher
            # mode
            self.keep_order_lock = threading.RLock()

            self.coordinator_thread = start_thread(self._coordinate)
        else:
            self.queue_of_output_q = None

        # `dispatcher` is a user-defined function to distribute args to workers.
        # It accepts the same args passed to worker and returns a number to
        # indicate which worker to used.
        if self.dispatcher is not None:
            self.dispatch_queues = []
            for i in range(self.n_thread):
                self.dispatch_queues.append({
                    'input': _make_q(),
                    'output': _make_q(),
                })
            self.dispatcher_thread = start_thread(self._dispatch)

        self.add_worker_thread()

    def add_worker_thread(self):

        with self.worker_group_lock:

            if self.exiting:
                logger.info('worker_group exiting.'
                            ' Thread number change not allowed')
                return

            s, e = self.running_index_range

            for i in range(s, e):

                if i not in self.threads:

                    if self.keep_order:
                        th = start_thread(self._exec_in_order,
                                          self.input_queue, _make_q(), i)
                    elif self.dispatcher is not None:
                        # for worker group with dispatcher, threads are added for the first time
                        assert s == 0
                        assert e == self.n_thread
                        th = start_thread(self._exec,
                                          self.dispatch_queues[i]['input'],
                                          self.dispatch_queues[i]['output'],
                                          i)
                    else:
                        th = start_thread(self._exec,
                                          self.input_queue, self.output_queue, i)

                    self.threads[i] = th

    def _exec(self, input_q, output_q, thread_index):

        while self.running:

            # If this thread is not in the running thread range, exit.
            if thread_index < self.running_index_range[0]:

                with self.worker_group_lock:
                    del self.threads[thread_index]

                logger.info('worker-thread {i} quit'.format(i=thread_index))
                return

            args = input_q.get()
            if args is Finish:
                return

            with self.probe['probe_lock']:
                self.probe['in'] += 1

            try:
                rst = self.worker(args)
            except Exception as e:
                logger.exception(repr(e))
                continue

            finally:
                with self.probe['probe_lock']:
                    self.probe['out'] += 1

            # If rst is an iterator, it procures more than one args to next job.
            # In order to be accurate, we only count an iterator as one.

            _put_rst(output_q, rst)

    def _exec_in_order(self, input_q, output_q, thread_index):

        while self.running:

            if thread_index < self.running_index_range[0]:

                with self.worker_group_lock:
                    del self.threads[thread_index]

                logger.info('in-order worker-thread {i} quit'.format(
                    i=thread_index))
                return

            with self.keep_order_lock:

                args = input_q.get()
                if args is Finish:
                    return
                self.queue_of_output_q.put(output_q)

            with self.probe['probe_lock']:
                self.probe['in'] += 1

            try:
                rst = self.worker(args)

            except Exception as e:
                logger.exception(repr(e))
                output_q.put(EmptyRst)
                continue

            finally:
                with self.probe['probe_lock']:
                    self.probe['out'] += 1

            output_q.put(rst)

    def _coordinate(self):

        while self.running:

            outq = self.queue_of_output_q.get()
            if outq is Finish:
                return

            _put_rst(self.output_queue, outq.get())

    def _dispatch(self):

        while self.running:

            args = self.input_queue.get()
            if args is Finish:
                return

            n = self.dispatcher(args)
            n = n % self.n_thread

            queues = self.dispatch_queues[n]
            inq, outq = queues['input'], queues['output']

            self.queue_of_output_q.put(outq)
            inq.put(args)


class JobManager(object):

    def __init__(self, workers, queue_size=1024, probe=None, keep_order=False):

        if probe is None:
            probe = {}

        self.workers = workers
        self.head_queue = _make_q(queue_size)
        self.probe = probe
        self.keep_order = keep_order

        self.worker_groups = []

        self.probe.update({
            'worker_groups': self.worker_groups,
            'probe_lock': threading.RLock(),
            'in': 0,
            'out': 0,
        })

        self.make_worker_groups()

    def make_worker_groups(self):

        inq = self.head_queue

        workers = self.workers + [_blackhole]
        for i, worker in enumerate(workers):

            if callable(worker):
                worker = (worker, 1)

            # worker callable, n_thread, dispatcher
            worker = (worker + (None,))[:3]

            worker, n, dispatcher = worker

            wg = WorkerGroup(i, worker, n, inq,
                             dispatcher,
                             self.probe, self.keep_order)

            self.worker_groups.append(wg)
            inq = wg.output_queue

    def set_thread_num(self, worker, n):

        # When thread number is increased, new threads are created.
        # If thread number is reduced, we do not stop worker thread in this
        # function.
        # Because we do not have a steady way to shutdown a running thread in
        # python.
        # Worker thread checks if it should continue running in _exec() and
        # _exec_in_order(), by checking its thread_index against running thread
        # index range running_index_range.

        assert(n > 0)
        assert(isinstance(n, int))

        for wg in self.worker_groups:

            """
            In python2, `x = X(); x.meth is x.meth` results in a `False`.
            Every time to retrieve a method, python creates a new **bound** function.

            We must use == to test function equality.

            See https://stackoverflow.com/questions/15977808/why-dont-methods-have-reference-equality
            """

            if wg.worker != worker:
                continue

            if wg.dispatcher is not None:
                raise JobWorkerError('worker-group with dispatcher does not allow to change thread number')

            with wg.worker_group_lock:

                if wg.exiting:
                    logger.info('worker group exiting.'
                                ' Thread number change not allowed')
                    break

                s, e = wg.running_index_range
                oldn = e - s

                if n < oldn:
                    s += oldn - n
                elif n > oldn:
                    e += n - oldn
                else:
                    break

                wg.running_index_range = [s, e]
                wg.add_worker_thread()

                logger.info('thread number is set to {n},'
                            ' thread index: {idx},'
                            ' running threads: {ths}'.format(
                                n=n,
                                idx=range(wg.running_index_range[0],
                                          wg.running_index_range[1]),
                                ths=sorted(wg.threads.keys())))
                break

        else:
            raise JobWorkerNotFound(worker)

    def put(self, elt):
        self.head_queue.put(elt)

    def join(self, timeout=None):

        endtime = time.time() + (timeout or 86400 * 365)

        for wg in self.worker_groups:

            with wg.worker_group_lock:
                # prevent adding or removing thread
                wg.exiting = True
                ths = wg.threads.values()

            if wg.dispatcher is None:
                # put nr = len(threads) Finish
                for th in ths:
                    wg.input_queue.put(Finish)
            else:
                wg.input_queue.put(Finish)
                # wait for dispatcher to finish or jobs might be lost
                wg.dispatcher_thread.join(endtime - time.time())

                for qs in wg.dispatch_queues:
                    qs['input'].put(Finish)

            for th in ths:
                th.join(endtime - time.time())

            if wg.queue_of_output_q is not None:
                wg.queue_of_output_q.put(Finish)
                wg.coordinator_thread.join(endtime - time.time())

            # if join timeout, let threads quit at next loop
            wg.running = False

    def stat(self):
        return stat(self.probe)


def run(input_it, workers, keep_order=False, timeout=None, probe=None):

    mgr = JobManager(workers, probe=probe, keep_order=keep_order)

    for args in input_it:
        mgr.put(args)

    mgr.join(timeout=timeout)


def stat(probe):

    with probe['probe_lock']:
        rst = {
            'in': probe['in'],
            'out': probe['out'],
            'doing': probe['in'] - probe['out'],
            'workers': [],
        }

    # exclude the _start and _end
    for wg in probe['worker_groups'][:-1]:
        o = {}
        wk = wg.worker
        o['name'] = (wk.__module__ or 'builtin') + ":" + wk.__name__
        o['input'] = _q_stat(wg.input_queue)
        if wg.dispatcher is not None:
            o['dispatcher'] = [
                {'input': _q_stat(qs['input']),
                 'output': _q_stat(qs['output']),
                 }
                for qs in wg.dispatch_queues
            ]

        if wg.queue_of_output_q is not None:
            o['coordinator'] = _q_stat(wg.queue_of_output_q)

        s, e = wg.running_index_range
        o['nr_worker'] = e - s

        rst['workers'].append(o)

    return rst


def _q_stat(q):
    return {'size': q.qsize(),
            'capa': q.maxsize
            }


def _put_rst(output_q, rst):

    if type(rst) == types.GeneratorType:
        for rr in rst:
            _put_non_empty(output_q, rr)
    else:
        _put_non_empty(output_q, rst)


def _blackhole(args):
    return EmptyRst


def _put_non_empty(q, val):
    if val is not EmptyRst:
        q.put(val)


def _make_q(n=1024):
    return Queue.Queue(n)


def start_thread(exec_func, *args):
    # `thread_index` identifying a thread in worker_group.threads.
    # It is used to decide which thread to remove.
    return threadutil.start_daemon_thread(exec_func, args=args)



import logging
import sys
import threading

if sys.version_info > (3,):
    import queue as Queue
else:
    import Queue

logger = logging.getLogger(__name__)


class CachePoolError(Exception):
    pass


class CachePoolGeneratorError(CachePoolError):
    pass


class CachePool(object):

    def __init__(self,
                 generator,
                 generator_args=None,
                 generator_argkw=None,
                 close_callback=None,
                 pool_size=1024):

        self._generator = generator
        self._generator_args = [] if generator_args is None else generator_args
        self._generator_argkw = {} if generator_argkw is None else generator_argkw

        self._close_callback = close_callback

        self._pool_size = pool_size
        self.queue = Queue.Queue(self._pool_size)

        self._stat_lock = threading.RLock()
        self.stat = {
            'new': 0,
            'get': 0,
            'put': 0,
        }

        if not callable(self._generator):
            raise CachePoolGeneratorError('generator is not callable')

    def get(self):

        try:
            element = self.queue.get(block=False)
            with self._stat_lock:
                self.stat['get'] += 1

            logger.debug('reuse: ' + repr(self.stat))
            logger.debug('qzise: ' + repr(self.queue.qsize()))

            return element

        except Queue.Empty:

            element = self._generator(
                *self._generator_args, **self._generator_argkw)

            with self._stat_lock:
                self.stat['new'] += 1
                self.stat['get'] += 1

            logger.debug('new  : ' + repr(self.stat))

            return element

    def put(self, element):

        try:
            self.queue.put(element, block=False)
            with self._stat_lock:
                self.stat['put'] += 1

            logger.debug('put  : ' + repr(self.stat))
            logger.debug('qzise: ' + repr(self.queue.qsize()))

        except Queue.Full:
            self.close(element)

    def close(self, element):

        if ((self._close_callback is not None)
                and callable(self._close_callback)):
            self._close_callback(element)


class CacheWrapper(object):

    def __init__(self, pool, reuse_decider=None):
        self.pool = pool
        self.element = None

        self._reuse_decider = reuse_decider

    def __enter__(self):
        self.element = self.pool.get()
        return self.element

    def __exit__(self, errtype, errval, _traceback):

        if errtype is None:
            self.pool.put(self.element)
        else:
            if ((self._reuse_decider is not None)
                    and callable(self._reuse_decider)
                    and self._reuse_decider(errtype, errval, _traceback)):

                self.pool.put(self.element)

            else:
                self.pool.close(self.element)

        self.element = None


def make_wrapper(pool, reuse_decider=None):

    def _wrapper():
        return CacheWrapper(pool, reuse_decider=reuse_decider)

    return _wrapper



#!/bin/env python2
# coding: utf-8

import copy
import operator
from collections import defaultdict


def depth_iter(mydict, ks=None, maxdepth=10240,
               intermediate=False, empty_leaf=False,
               is_allowed=None):

    ks = ks or []

    dickeys = sorted(mydict.keys())
    for k in dickeys:

        v = mydict[k]

        ks.append(k)

        if len(ks) >= maxdepth:
            if is_allowed is None or is_allowed(ks, v):
                yield ks, v
        else:
            if isinstance(v, dict):

                if is_allowed is not None:
                    if is_allowed(ks, v):
                        yield ks, v
                else:
                    if intermediate or (empty_leaf and len(v) == 0):
                        yield ks, v

                for _ks, v in depth_iter(v, ks,
                                         maxdepth=maxdepth,
                                         intermediate=intermediate,
                                         empty_leaf=empty_leaf,
                                         is_allowed=is_allowed,
                                         ):
                    yield _ks, v
            else:
                if is_allowed is None or is_allowed(ks, v):
                    yield ks, v

        ks.pop(-1)


def breadth_iter(mydict):

    q = [([], mydict)]

    while True:
        if len(q) < 1:
            break

        _ks, node = q.pop(0)
        for k, v in node.items():
            ks = _ks[:]
            ks.append(k)
            yield ks, v

            if isinstance(v, dict):
                q.append((ks, v))


def get(dic, key_path, vars=None, default=0, ignore_vars_key_error=None):

    if vars is None:
        vars = {}

    if ignore_vars_key_error is None:
        ignore_vars_key_error = True

    _default = vars.get('_default', default)
    node = dic

    _keys = key_path

    if isinstance(key_path, basestring):
        _keys = key_path.split('.')

    for k in _keys:

        try:
            key = _translate_var(k, vars)
        except KeyError:
            if ignore_vars_key_error:
                return _default
            else:
                raise

        if key not in node:
            return _default

        node = node[key]

    return node


def make_getter_str(key_path, default=0):

    s = 'lambda dic, vars={}: dic'

    _keys = key_path
    if isinstance(key_path, basestring):
        _keys = key_path.split('.')

    for k in _keys:

        k_str = _translate_var_str(k)

        s += '.get(%s, {})' % (k_str, )

    s = s[:-3] + 'vars.get("_default", ' + repr(default) + '))'

    return s


def _translate_var(k, vars):
    if isinstance(k, basestring):
        if k.startswith('$'):
            k = k[1:]
            if k in vars:
                return str(vars[k])
            else:
                raise KeyError('{k} does not exist in vars: {vars}'.format(
                    k=k, vars=vars))
        else:
            return k
    elif isinstance(k, tuple):

        return tuple(_translate_var(kk, vars) for kk in k)

    else:
        return k


def _translate_var_str(k):
    if isinstance(k, basestring):
        if k.startswith('$'):
            return 'str(vars.get("%s", "_"))' % (k[1:],)
        else:
            return '"' + k + '"'

    elif isinstance(k, tuple):
        s = "("
        for kk in k:
            s += _translate_var_str(kk) + ','
        return s + ")"
    else:
        return repr(k)


def make_getter(key_path, default=0):
    return eval(make_getter_str(key_path, default=default))


def make_setter(key_path, value=None, incr=False):
    _keys = key_path
    if isinstance(key_path, basestring):
        _keys = key_path.split('.')

    def_val = value

    def _set_dict(dic, value=None, vars={}):

        k = 'self'
        _node = {'self': dic}

        for _k in _keys:

            if k not in _node:
                _node[k] = {}
            _node = _node[k]

            k = _translate_var(_k, vars)

        if value is not None:
            val_to_set = value
        else:
            val_to_set = def_val

        if callable(val_to_set):
            val_to_set = val_to_set(vars)

        if k not in _node:
            # use the default constructor to get a default "zero" value
            _node[k] = type(val_to_set)()

        if incr:
            _node[k] += val_to_set
        else:
            _node[k] = val_to_set

        return _node[k]

    return _set_dict


def _contains(a, b, ref_table):
    if a is b:
        return True

    if (isinstance(a, list) and isinstance(b, list)
            or (isinstance(a, tuple) and isinstance(b, tuple))):

        if len(a) < len(b):
            return False

        for i, v in enumerate(b):
            if not _contains(a[i], v, ref_table):
                return False
        else:
            return True

    if not isinstance(a, dict) or not isinstance(b, dict):
        return a == b

    id_a, id_b = id(a), id(b)

    if ref_table[id_a].get(id_b) is not None:
        return ref_table[id_a][id_b]

    ref_table[id_a][id_b] = True

    for k, v in b.items():
        if a.get(k) is None:
            return False

        if not _contains(a[k], v, ref_table):
            return False

    return True


def contains(a, b):
    return _contains(a, b, defaultdict(dict))


class AttrDict(dict):

    '''
    a = AttrDict({1:2}) # {1:2}
    a = AttrDict(x=3)   # {"x":3}
    a.x                 # 3
    a['x']              # 3

    Some pros:

    - It actually works!
    - No dictionary class methods are shadowed (e.g. .keys() work just fine)
    - Attributes and items are always in sync
    - Trying to access non-existent key as an attribute correctly raises AttributeError instead of KeyError

    Cons:

    - Methods like .keys() will not work just fine if they get overwritten by incoming data
    - Causes a memory leak in Python < 2.7.4 / Python3 < 3.2.3
    - Pylint goes bananas with E1123(unexpected-keyword-arg) and E1103(maybe-no-member)
    - For the uninitiated it seems like pure magic.

    Issues:

    - Dictionary key overrides dictionary methods:

      d = AttrDict()
      d.update({'items':["a", "b"]})
      d.items() # TypeError: 'list' object is not callable

    '''

    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self


class AttrDictCopy(dict):

    # Allow to set attribute or key.
    # But when get attribute or key, the value is copied before returning.
    # To prevent changing original data.

    def __getattr__(self, k):

        if k not in self:
            raise AttributeError(repr(k) + ' not found')

        return self[k]

    def __setattr__(self, k, v):
        raise AttributeError('AttrDictCopy does not allow to set attribute')

    def __getitem__(self, k):

        if k not in self:
            raise KeyError(repr(k) + ' not found')

        v = super(AttrDictCopy, self).__getitem__(k)
        if isinstance(v, AttrDictCopy):
            # reduce it to a normal dict, or deepcopy can not set items to the new instance
            v = v.as_dict()
            v = copy.deepcopy(v)
            return _attrdict(AttrDictCopy, v, {})
        else:
            return copy.deepcopy(v)

    def __setitem__(self, k, v):
        raise KeyError('AttrDictCopy does not allow to set key')

    def as_dict(self):
        d = {}

        for k in self.keys():
            v = super(AttrDictCopy, self).__getitem__(k)
            if isinstance(v, AttrDictCopy):
                v = v.as_dict()

            d[k] = v

        return d


def attrdict(*args, **kwargs):
    """
    Make a dict-like object whose keys can also be accessed with attribute.
    You can use an AttrDict instance just like using a dict instance.
    """

    d = dict(*args, **kwargs)
    return _attrdict(AttrDict, d, {})


def attrdict_copy(*args, **kwargs):

    d = dict(*args, **kwargs)
    return _attrdict(AttrDictCopy, d, {})


def _attrdict(attrdict_clz, d, ref):

    if not isinstance(d, dict):
        return d

    if isinstance(d, attrdict_clz):
        return d

    if id(d) in ref:
        return ref[id(d)]

    # id() is the memory address of an object, thus it is unique.
    ad = attrdict_clz(d)
    ref[id(d)] = ad

    for k in d.keys():
        sub_ad = _attrdict(attrdict_clz, d[k], ref)
        super(attrdict_clz, ad).__setitem__(k, sub_ad)

    return ad


def combineto(a, b, op, exclude=None, recursive=True):

    if not isinstance(b, dict):
        return a

    if exclude is None:
        exclude = {}

    for k, vb in b.iteritems():
        sub_exclude = exclude.get(k)

        if sub_exclude is True:
            continue

        if k not in a:
            # use the default constructor of `vb` to get a `zero` value.
            va = type(vb)()
        else:
            va = a[k]

        if isinstance(vb, dict):
            if recursive:
                a[k] = combineto(va, vb, op, exclude=sub_exclude, recursive=recursive)
            else:
                continue
        else:
            a[k] = op(va, vb)

    return a


def combine(a, b, op, exclude=None, recursive=True):
    r = copy.deepcopy(a)
    combineto(r, b, op, exclude=exclude, recursive=recursive)
    return r


def addto(a, b, exclude=None, recursive=True):
    op = operator.add
    return combineto(a, b, op, exclude=exclude, recursive=recursive)


def add(a, b, exclude=None, recursive=True):
    r = copy.deepcopy(a)
    addto(r, b, exclude=exclude, recursive=recursive)
    return r


def subdict(source, flds, use_default=False, default=None, deepcopy=False, deepcopy_default=False):

    result = {}

    for k in flds:

        if k in source:
            val = source[k]
            result[k] = _copy_value(val, deepcopy)
            continue

        if use_default is False:
            continue

        if callable(default):
            result[k] = default(k)
            continue

        result[k] = _copy_value(default, deepcopy_default)

    return result


def _copy_value(val, deepcopy=False):

    if deepcopy:
        return copy.deepcopy(val)

    return val



#!/usr/bin/env python
# coding: utf-8

import logging

from pykit import rangeset
from pykit import txutil

from .accessor import KeyValue
from .accessor import Value
from .status import STATUS

logger = logging.getLogger(__name__)


class StorageHelper(object):

    # keeps the last n modifications in a record
    max_value_history = 16
    conflicterror = None

    def apply_record(self, txid, key, value):

        # the data in underlying storage is multi-version record:
        # [
        #     [<txid>, <value>]
        #     [<txid>, <value>]
        #     ...
        # ]

        for curr in txutil.cas_loop(self.record.get,
                                    self.record.set_or_create,
                                    args=(key, ),
                                    conflicterror=self.conflicterror):

            max_txid = curr.v[-1][0]

            if max_txid >= txid:
                return False

            curr.v.append((txid, value))
            while len(curr.v) > self.max_value_history:
                curr.v.pop(0)

        return True

    def add_to_txidset(self, status, txid):

        if status not in STATUS:
            raise KeyError('invalid status: ' + repr(status))

        logger.info('add {status}:{txid}'
                    ' to txidset'.format(
                        status=status, txid=txid))

        for curr in txutil.cas_loop(self.txidset.get,
                                    self.txidset.set,
                                    conflicterror=self.conflicterror):

            for st in STATUS:
                if st not in curr.v:
                    curr.v[st] = rangeset.RangeSet([])

            curr.v[status].add([txid, txid + 1])


class Storage(StorageHelper):

    record = KeyValue()
    journal = KeyValue()
    txidset = Value()

    def acquire_key_loop(self, txid, key): raise TypeError('unimplemented')

    def try_release_key(self, txid, key): raise TypeError('unimplemented')



#######################
# unit test case read this file
import os
import sys

fd = int(sys.argv[1])

try:
    a = os.read(fd, 3)
    print a
    sys.exit(0)
except OSError as e:
    print 'errno=' + str(e.errno)
    sys.exit(1)



#!/usr/bin/env python
# coding: utf-8

import kazoo.exceptions


class TXError(Exception):
    pass


class Aborted(TXError):
    pass


class NotLocked(Aborted):
    pass


class UnlockNotAllowed(Aborted):
    pass


class RetriableError(TXError):
    pass


class UserAborted(Aborted):
    pass


class HigherTXApplied(Aborted, RetriableError):
    pass


class Deadlock(Aborted, RetriableError):
    pass


class TXTimeout(Aborted):
    pass


class ConnectionLoss(TXError, kazoo.exceptions.ConnectionLoss):
    pass



#!/usr/bin/env python
# coding: utf-8

# https://dev.mysql.com/doc/refman/5.7/en/privileges-provided.html

# ALL [PRIVILEGES]                Synonym for “all privileges”    Server administration
# ALTER                           Alter_priv                      Tables
# ALTER ROUTINE                   Alter_routine_priv              Stored routines
# CREATE                          Create_priv                     Databases, tables, or indexes
# CREATE ROUTINE                  Create_routine_priv             Stored routines
# CREATE TABLESPACE               Create_tablespace_priv          Server administration
# CREATE TEMPORARY TABLES         Create_tmp_table_priv           Tables
# CREATE USER                     Create_user_priv                Server administration
# CREATE VIEW                     Create_view_priv                Views
# DELETE                          Delete_priv                     Tables
# DROP                            Drop_priv                       Databases, tables, or views
# EVENT                           Event_priv                      Databases
# EXECUTE                         Execute_priv                    Stored routines
# FILE                            File_priv                       File access on server host
# GRANT OPTION                    Grant_priv                      Databases, tables, or stored routines
# INDEX                           Index_priv                      Tables
# INSERT                          Insert_priv                     Tables or columns
# LOCK TABLES                     Lock_tables_priv                Databases
# PROCESS                         Process_priv                    Server administration
# PROXY                           See proxies_priv table          Server administration
# REFERENCES                      References_priv                 Databases or tables
# RELOAD                          Reload_priv                     Server administration
# REPLICATION CLIENT              Repl_client_priv                Server administration
# REPLICATION SLAVE               Repl_slave_priv                 Server administration
# SELECT                          Select_priv                     Tables or columns
# SHOW DATABASES                  Show_db_priv                    Server administration
# SHOW VIEW                       Show_view_priv                  Views
# SHUTDOWN                        Shutdown_priv                   Server administration
# SUPER                           Super_priv                      Server administration
# TRIGGER                         Trigger_priv                    Tables
# UPDATE                          Update_priv                     Tables or columns
# USAGE                           Synonym for “no privileges”     Server administration

privileges = {
        'ALL':                      ('ALL',                     ),
        'ALTER':                    ('ALTER',                   ),
        'ALTER_ROUTINE':            ('ALTER ROUTINE',           ),
        'CREATE':                   ('CREATE',                  ),
        'CREATE_ROUTINE':           ('CREATE ROUTINE',          ),
        'CREATE_TABLESPACE':        ('CREATE TABLESPACE',       ),
        'CREATE_TEMPORARY_TABLES':  ('CREATE TEMPORARY TABLES', ),
        'CREATE_USER':              ('CREATE USER',             ),
        'CREATE_VIEW':              ('CREATE VIEW',             ),
        'DELETE':                   ('DELETE',                  ),
        'DROP':                     ('DROP',                    ),
        'EVENT':                    ('EVENT',                   ),
        'EXECUTE':                  ('EXECUTE',                 ),
        'FILE':                     ('FILE',                    ),
        'GRANT_OPTION':             ('GRANT OPTION',            ),
        'INDEX':                    ('INDEX',                   ),
        'INSERT':                   ('INSERT',                  ),
        'LOCK_TABLES':              ('LOCK TABLES',             ),
        'PROCESS':                  ('PROCESS',                 ),
        'PROXY':                    ('PROXY',                   ),
        'REFERENCES':               ('REFERENCES',              ),
        'RELOAD':                   ('RELOAD',                  ),
        'REPLICATION_CLIENT':       ('REPLICATION CLIENT',      ),
        'REPLICATION_SLAVE':        ('REPLICATION SLAVE',       ),
        'SELECT':                   ('SELECT',                  ),
        'SHOW_DATABASES':           ('SHOW DATABASES',          ),
        'SHOW_VIEW':                ('SHOW VIEW',               ),
        'SHUTDOWN':                 ('SHUTDOWN',                ),
        'SUPER':                    ('SUPER',                   ),
        'TRIGGER':                  ('TRIGGER',                 ),
        'UPDATE':                   ('UPDATE',                  ),
        'USAGE':                    ('USAGE',                   ),
}

# set direct map

for k, v in privileges.items():
    privileges[v[0]] = v

# predefined shortcuts

privileges.update({
        'replicator': (  privileges["REPLICATION_CLIENT"]
                       + privileges["REPLICATION_SLAVE"]
                       + privileges["SELECT"]
        ),
        'monitor':    (  privileges["SELECT"]
        ),
        'business':   (  privileges["CREATE"]
                       + privileges["DROP"]
                       + privileges["REFERENCES"]
                       + privileges["ALTER"]
                       + privileges["DELETE"]
                       + privileges["INDEX"]
                       + privileges["INSERT"]
                       + privileges["SELECT"]
                       + privileges["UPDATE"]
        ),
        'readwrite':  (  privileges["DELETE"]
                       + privileges["INSERT"]
                       + privileges["SELECT"]
                       + privileges["UPDATE"]
        ),
})



#!/usr/bin/env python2
# coding: utf-8

import os
import unittest

from pykit import dictutil
from pykit import strutil
from pykit import ututil

dd = ututil.dd

this_base = os.path.dirname(__file__)

class TestTrie(unittest.TestCase):

    def setUp(self):
        self.strs = (
                'abc',
                'abcdef',
                'abcdeg',
                'abcdfg',
                'abd',
                'abe',
                'axy',
                'b11',
                'b12',
                'b123',
                'b14',
        )
        self.iterables = (
            (1, 'ab', '我'),
            (1, 'ab', '我是'),
            (1, 'ab', '我是好'),
            (1, 'ab', '我是好人'),
            (1, 'c',  '我是好人'),
            (1, 'd',  '人'),
            (1, 'd',  '人人'),
            (1, 'd',  '人人人'),
            (1, 'd',  '人人人人'),
            (2, 'a',  '人人'),
            (2, 'b',  'x'),
            (2, 'b',  'x', 'a'),
            (2, 'b',  'x', 'b'),
            (2, 'b',  'x', 'c'),
            (2, 'b',  'x', 'cd'),
            (2, 'b',  'xx'),
            (2, 'b',  'y'),
            (2, 'b',  'z'),
        )

    def test_trie_iterable(self):

        t = strutil.make_trie(self.iterables, node_max_num=3)

        dd('trie:')
        dd(str(t))

        self.assertEqual(len(self.iterables), t.n)

        for ks, v in dictutil.depth_iter(t, is_allowed=lambda ks, v: v.is_eol):

            s = tuple(ks)
            dd('whole string: ', s, v)
            self.assertTrue(s in self.iterables)

    def test_sharding_iterable(self):

        dd()

        expected = (
            (None,           5),
            ((1, 'd', '人'), 5),
            ((2, 'b', 'x'),  5),
            ((2, 'b', 'xx'), 3),
        )

        rst = strutil.sharding(self.iterables, size=5, joiner=tuple)
        for i, (start, n) in enumerate(rst):
            dd('{start:<20} {n:>10}'.format(start=start, n=n))
            self.assertEqual(expected[i], (start, n))

    def test_sharding_accuracy(self):

        dd()

        cases = (
                [ 5,
                  ((None,           5),
                   ((1, 'd',    ),  5),
                   ((2, 'b', 'x'),  5),
                   ((2, 'b', 'xx'), 3),
                  ),
                ],
                [ 9,
                  ((None,  9),
                   ((2,),  9),
                  ),
                ],
                [ len(self.iterables),
                  ((None,  len(self.iterables)),
                  ),
                ],
                [ len(self.iterables)+1,
                  ((None,  len(self.iterables)),
                  ),
                ],
        )

        for accuracy, expected in cases:
            rst = strutil.sharding(self.iterables, size=5,
                                   accuracy=accuracy, joiner=tuple)

            dd("accuracy:", accuracy)
            dd("rst:")
            dd(rst)

            for i, (start, n) in enumerate(rst):
                dd('{start:<20} {n:>10}'.format(start=start, n=n))
                self.assertEqual(expected[i], (start, n))

    def test_trie_whole_string(self):

        t = strutil.make_trie(self.strs, node_max_num=3)

        dd('trie:')
        dd(str(t))

        self.assertEqual(len(self.strs), t.n)

        # only whole strings
        for ks, v in dictutil.depth_iter(t, is_allowed=lambda ks, v: v.is_eol):

            s = ''.join(ks)
            dd('whole string: ', s, v)
            self.assertTrue(s in self.strs)

    def test_trie_prefix_count(self):

        t = strutil.make_trie(self.strs, node_max_num=3)

        dd('==== trie:')
        dd(str(t))

        # all prefixes, might be also a whole string.
        for ks, v in dictutil.depth_iter(t, intermediate=True):

            dd('==== got keys:', ks, 'sub trie:')
            dd(v)

            prefix = ''.join(ks)
            expected = len([x
                            for x in self.strs
                            if x.startswith(prefix)])

            key_path = '.'.join(ks)
            subtrie = dictutil.get(t, key_path, default=-1)
            rst = subtrie.n

            dd('prefix:', prefix, 'trie key_path:', key_path, 'sub trie:')
            dd(str(subtrie))
            dd('expected:', expected)
            dd('rst:', rst)

            self.assertEqual(expected, rst)

            # Leaf node number limit
            if len(v) == 0:
                self.assertLessEqual(rst, 3)


    def test_node_max_num(self):

        t = strutil.make_trie(self.strs, node_max_num=1)

        for ks, v in dictutil.depth_iter(t, empty_leaf=True):

            key_path = '.'.join(ks)
            subtrie = dictutil.get(t, key_path, default=-1)

            self.assertEqual(1, subtrie.n)

        # all in one node

        t = strutil.make_trie(self.strs, node_max_num=10000)

        dd()
        dd(str(t))

        self.assertEqual(len(self.strs), t.n)
        self.assertEqual(0, len(t))

    def test_sharding(self):

        with open(os.path.join(this_base, 'words')) as f:
            lines = f.readlines()
            lines = [x.strip() for x in lines]

        _size, _accuracy = 200, 20

        rst = strutil.sharding(lines, size=_size, accuracy=_accuracy)

        expected = [
            (None,    209, ),
            ('M',     202, ),
            ('TestU', 202, ),
            ('br',    202, ),
            ('dc',    201, ),
            ('exi',   202, ),
            ('inf',   204, ),
            ('may',   205, ),
            ('pf',    200, ),
            ('rew',   208, ),
            ('suc',   204, ),
            ('wh',    56, ),
        ]

        for i, (start, size) in enumerate(rst):
            dd('{start:<10} {size:>20}'.format(start=start, size=size))
            self.assertEqual(expected[i], (start, size))

        # general cases

        cases = (
            (1, 1),
            (20, 1),
            (20, 5),
            (200, 20),
            (200, 100),
        )

        for _size, _accuracy in cases:

            dd('size: {_size}, accuracy: {_accuracy}'.format(_size=_size, _accuracy=_accuracy))

            rst = strutil.sharding(lines, size=_size, accuracy=_accuracy)

            tot = 0
            for i, (start, size) in enumerate(rst):

                dd('{start:<10} {size:>20}'.format(start=start, size=size))

                # the last shard might have less items
                if i < len(rst) - 1:
                    nxt = rst[i+1][0]
                    self.assertLessEqual(_size, size)
                    self.assertLessEqual(size, _size + _accuracy)
                    self.assertEqual(len([x for x in lines
                                          if x >= start and x < nxt ]),
                                     size)
                else:
                    self.assertLessEqual(size, _size)
                    self.assertEqual(len([x for x in lines
                                          if x >= start ]),
                                     size)

                tot += size

            self.assertEqual(len(lines), tot)



from .checker import (
    InvalidArgumentError,
    InvalidLengthError,
    InvalidSchemaError,
    InvalidTypeError,
    InvalidValueError,
    LackArgumentError,
    NotInEnumError,
    OutOfRangeError,
    PatternNotMatchError,
    SchemaCheckerError,
    SchemaNotFoundError,
    UnexpectedArgumentError,

    check_arguments,
    read_schema,
    validate_schema,
)

__all__ = [
    "InvalidArgumentError",
    "InvalidLengthError",
    "InvalidSchemaError",
    "InvalidTypeError",
    "InvalidValueError",
    "LackArgumentError",
    "NotInEnumError",
    "OutOfRangeError",
    "PatternNotMatchError",
    "SchemaCheckerError",
    "SchemaNotFoundError",
    "UnexpectedArgumentError",

    "check_arguments",
    "read_schema",
    "validate_schema",
]



#!/usr/bin/env python2
# coding: utf-8
import logging
from pykit import wsjobd

PORT = 33445


def run():
    wsjobd.run(ip='127.0.0.1', port=PORT, jobq_thread_count=20)


if __name__ == "__main__":
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)

    file_handler = logging.FileHandler('wsjobd.log')
    formatter = logging.Formatter('[%(asctime)s, %(levelname)s] %(message)s')
    file_handler.setFormatter(formatter)
    logger.addHandler(file_handler)

    from pykit import daemonize
    daemonize.daemonize_cli(run, '/tmp/wsjod_server.pid')



#!/usr/bin/env python2
# coding: utf-8

import os
import time
import unittest

from pykit import daemonize
from pykit import proc
from pykit import ututil

dd = ututil.dd

this_base = os.path.dirname(__file__)


def subproc(script, env=None):
    if env is None:
        env = dict(PYTHONPATH=this_base + '/../..',)

    return proc.shell_script(script, env=env)


def read_file(fn):
    try:
        with open(fn, 'r') as f:
            cont = f.read()
            return cont
    except EnvironmentError:
        return None


class TestDaemonize(unittest.TestCase):

    foo_fn = '/tmp/foo'
    bar_fn = '/tmp/bar'
    pidfn = '/tmp/test_daemonize.pid'

    def _clean(self):

        # kill foo.py and kill bar.py
        # bar.py might be waiting for foo.py to release lock-file.
        try:
            subproc('python2 {b}/foo.py stop'.format(b=this_base))
        except Exception as e:
            dd(repr(e))

        time.sleep(0.1)

        try:
            subproc('python2 {b}/bar.py stop'.format(b=this_base))
        except Exception as e:
            dd(repr(e))

        # remove written file

        try:
            os.unlink(self.foo_fn)
        except EnvironmentError as e:
            pass

        try:
            os.unlink(self.bar_fn)
        except EnvironmentError as e:
            pass

    def setUp(self):
        self._clean()

    def tearDown(self):
        self._clean()

    def test_start(self):

        subproc('python2 {b}/foo.py start'.format(b=this_base))
        time.sleep(0.2)

        self.assertEqual('foo-before', read_file(self.foo_fn))
        time.sleep(1)
        self.assertEqual('foo-after', read_file(self.foo_fn))

    def test_stop(self):

        subproc('python2 {b}/foo.py start'.format(b=this_base))
        time.sleep(0.2)

        self.assertEqual('foo-before', read_file(self.foo_fn), 'foo started')

        subproc('python2 {b}/foo.py stop'.format(b=this_base))
        time.sleep(0.2)

        self.assertEqual('foo-before', read_file(self.foo_fn),
                         'process has been kill thus no content is updated')

    def test_restart(self):

        subproc('python2 {b}/foo.py start'.format(b=this_base))
        time.sleep(0.2)

        self.assertEqual('foo-before', read_file(self.foo_fn))

        os.unlink(self.foo_fn)
        self.assertEqual(None, read_file(self.foo_fn))

        subproc('python2 {b}/foo.py restart'.format(b=this_base))
        time.sleep(0.2)

        self.assertEqual('foo-before', read_file(self.foo_fn),
                         'restarted and rewritten to the file')

    def test_exclusive_pid(self):

        subproc('python2 {b}/foo.py start'.format(b=this_base))
        time.sleep(0.1)
        subproc('python2 {b}/bar.py start'.format(b=this_base))
        time.sleep(0.1)

        self.assertEqual(None, read_file(self.bar_fn),
                         'bar.py not started or run')

    def test_default_pid_file(self):

        d = daemonize.Daemon()
        self.assertEqual('/var/run/__main__', d.pidfile)

    def test_close_fds(self):
        env = dict(PYTHONPATH='{path_daemonize}:{path_pykit}'.format(
                              path_daemonize=this_base + '/../..',
                              path_pykit=this_base + '/../../..'))

        subproc('python2 {b}/close_fds.py close'.format(b=this_base), env=env)
        time.sleep(0.2)

        fds = read_file(self.foo_fn)

        self.assertNotIn(self.bar_fn, fds)

        subproc('python2 {b}/close_fds.py open'.format(b=this_base), env=env)
        time.sleep(0.2)

        fds = read_file(self.foo_fn)

        self.assertIn(self.bar_fn, fds)



from pykit.cgrouparch import blkio
from pykit.cgrouparch import cpu

subsystem = {
    'cpu': {
        'set_cgroup': cpu.set_cgroup,
        'reset_statistics': cpu.reset_statistics,
        'account': cpu.account,
    },
    'blkio': {
        'set_cgroup': blkio.set_cgroup,
        'reset_statistics': blkio.reset_statistics,
        'account': blkio.account,
    },
}



#!/usr/bin/env python
# coding: utf-8

ABORTED = 'ABORTED'
COMMITTED = 'COMMITTED'
PURGED = 'PURGED'
PAUSED = 'PAUSED'
UNKNOWN = 'UNKNOWN'

STATUS = dict(ABORTED=ABORTED,
              COMMITTED=COMMITTED,
              PURGED=PURGED,
              )



#!/usr/bin/env python2
# coding: utf-8

import os
import time
import unittest

from pykit import fsutil
from pykit import proc
from pykit import threadutil
from pykit import ututil

dd = ututil.dd

# xx/pykit/fsutil/test/
this_base = os.path.dirname(__file__)

pyt = 'python2'


class TestCat(unittest.TestCase):

    fn = '/tmp/pykit-cat-test-content'

    def setUp(self):
        force_remove(self.fn)
        c = fsutil.Cat(self.fn)
        force_remove(c.stat_path())

    def tearDown(self):
        force_remove(self.fn)
        c = fsutil.Cat(self.fn)
        force_remove(c.stat_path())

    def test_entire_file(self):

        expected = [
            'a',
            'bcd',
        ]

        append_lines(self.fn, expected)

        rst = []
        for l in fsutil.Cat(self.fn, strip=True).iterate(timeout=0):
            rst.append(l)

        self.assertEqual(expected, rst)

    def test_entire_large_file(self):

        expected = [
            'a' * 1024 * 32,
            'b' * 1024 * 32,
        ] * 1024

        append_lines(self.fn, expected)

        dd(os.system('ls -l /tmp/pykit*'))

        rst = []
        for l in fsutil.Cat(self.fn, strip=True).iterate(timeout=0):
            rst.append(l)

        self.assertEqual(expected, rst)

    def test_half_line(self):

        n = 1024

        def _append():
            for i in range(n):
                time.sleep(0.001)
                append_bytes(self.fn, 'a' * 2048)
                time.sleep(0.001)
                append_bytes(self.fn, 'b' * 2048 + '\n')

        th = threadutil.start_daemon(_append)

        i = 0

        try:
            for l in fsutil.Cat(self.fn, strip=True).iterate(timeout=0.1):
                self.assertEqual('b', l[-1])
                i += 1
        except fsutil.NoData:
            pass

        th.join()
        self.assertEqual(n, i)

    def test_offset_record_when_destory(self):

        expected = [x * 32 for x in 'qwertyuiop']

        append_lines(self.fn, expected)
        cat_handle = fsutil.Cat(self.fn, strip=True)

        rst = []
        for val in cat_handle.iterate(timeout=0):
            rst.append(val)
        self.assertEqual(expected, rst)

        # stat file was removed
        fsutil.remove(cat_handle.stat_path())
        for val in cat_handle.iterate(timeout=0):
            rst.append(val)
        self.assertEqual(expected * 2, rst)

        # stat file was damaged
        fsutil.write_file(cat_handle.stat_path(), '{]')
        for val in cat_handle.iterate(timeout=0):
            rst.append(val)
        self.assertEqual(expected * 3, rst)

    def test_data_chucked(self):

        expected = ['a' * 32]
        new_data = ['b' * 32]
        chucked = ['c' * 31]

        cat_handle = fsutil.Cat(self.fn, strip=True)
        rst = []

        append_lines(self.fn, expected)
        for l in cat_handle.iterate(timeout=0):
            rst.append(l)
        self.assertEqual(expected, rst)

        # file was refreshed
        os.rename(self.fn, self.fn + '_old')
        append_lines(self.fn, new_data)
        fsutil.remove(self.fn + '_old')
        for l in cat_handle.iterate(timeout=0):
            rst.append(l)
        self.assertEqual(expected + new_data, rst)

        # file was chucked
        fsutil.write_file(self.fn, chucked[0])
        for l in cat_handle.iterate(timeout=0):
            rst.append(l)
        dd(rst)
        self.assertEqual(expected + new_data + chucked, rst)

    def test_continue_read(self):

        expected = [
            'a' * 32,
            'b' * 32,
        ]
        rst = []

        for i in xrange(1, 11):
            append_lines(self.fn, expected)

            for l in fsutil.Cat(self.fn, strip=True).iterate(timeout=0):
                rst.append(l)

            self.assertEqual(expected * i, rst)

    def test_file_change(self):

        expected = [
            'a' * 32,
            'b' * 32,
            'c' * 32,
            'd' * 32,
        ]
        rst = []

        def _override():
            for l in expected:
                time.sleep(0.1)
                force_remove(self.fn)
                append_lines(self.fn, [l])
                dd('overrided: ', l)

        th = threadutil.start_daemon(_override)

        try:
            for l in fsutil.Cat(self.fn, strip=True).iterate(timeout=0.4):
                rst.append(l)
        except fsutil.NoData:
            pass

        th.join()

        self.assertEqual(expected, rst)

    def test_wait_for_file(self):

        expected = [
            'a' * 32,
            'b' * 32,
        ]
        rst = []

        def _append():
            time.sleep(0.1)
            append_lines(self.fn, expected)
            dd('appended')

        th = threadutil.start_daemon(_append)

        try:
            for l in fsutil.Cat(self.fn, strip=True).iterate(timeout=0.4):
                rst.append(l)
        except fsutil.NoData:
            pass

        th.join()

        self.assertEqual(expected, rst)

    def test_wait_for_file_timeout(self):

        expected = [
            'a' * 32,
            'b' * 32,
        ]
        rst = []

        def _append():
            time.sleep(0.3)
            append_lines(self.fn, expected)
            dd('appended')

        th = threadutil.start_daemon(_append)

        try:
            for l in fsutil.Cat(self.fn, strip=True).iterate(timeout=0.1):
                rst.append(l)

            self.failed('expect NoSuchFile to raise')
        except fsutil.NoSuchFile:
            pass

        self.assertEqual([], rst)

        th.join()

    def test_wait_for_data(self):

        expected = [
            'a' * 32,
            'b' * 32,
        ]
        rst = []

        append_lines(self.fn, expected)

        def _append():
            for _ in range(5):
                time.sleep(0.1)
                append_lines(self.fn, expected)
            dd('appended')

        th = threadutil.start_daemon(_append)

        try:
            for l in fsutil.Cat(self.fn, strip=True).iterate(timeout=0.3):
                rst.append(l)

            self.failed('expect NoData to raise')
        except fsutil.NoData:
            pass

        th.join()
        self.assertEqual(expected * 6, rst)

    def test_wait_for_data_timeout(self):

        expected = [
            'a' * 32,
            'b' * 32,
        ]
        rst = []

        append_lines(self.fn, expected)
        dd('appended 1')

        def _append():
            time.sleep(0.3)
            append_lines(self.fn, expected)
            dd('appended')

        th = threadutil.start_daemon(_append)

        try:
            for l in fsutil.Cat(self.fn, strip=True).iterate(timeout=0.1):
                rst.append(l)

            self.failed('expect NoData to raise')
        except fsutil.NoData:
            pass

        th.join()
        self.assertEqual(expected, rst)

    def test_exclusive(self):

        append_lines(self.fn, ['x'])

        # by default it is exclusive
        a = fsutil.Cat(self.fn, strip=True).iterate(timeout=0.1)
        b = fsutil.Cat(self.fn, strip=True).iterate(timeout=0.1)

        val = a.next()
        self.assertEqual('x', val)
        self.assertRaises(fsutil.LockTimeout, b.next)

        # non-exclusive still work
        # a has not yet quit, thus it has not yet written offset to stat file.
        c = fsutil.Cat(self.fn, strip=True, exclusive=False).iterate(timeout=0.1)
        val = c.next()
        self.assertEqual('x', val)

    def test_handler(self):

        expected = [
            'a' * 32,
            'b' * 32,
        ]
        append_lines(self.fn, expected)

        rst = []
        c = fsutil.Cat(self.fn, strip=True, handler=rst.append)
        c.cat(timeout=0)
        self.assertEqual(expected, rst)

        force_remove(c.stat_path())

        # handler is a list of callable
        exp = []
        for l in expected:
            exp.append(l)
            exp.append(l)

        rst = []
        force_remove(self.fn)
        append_lines(self.fn, expected)
        c = fsutil.Cat(self.fn, strip=True, handler=[rst.append, rst.append])
        c.cat(timeout=0)

        self.assertEqual(exp, rst)

    def test_file_end_handler(self):
        expected = [
            'a' * 32,
            'b' * 32,
        ]
        append_lines(self.fn, expected)

        rst = []

        def _end():
            rst.append('end')

        # file_end_handler in cat()
        c = fsutil.Cat(self.fn, strip=True, handler=rst.append, file_end_handler=_end)
        c.cat(timeout=0)
        self.assertEqual(expected + ['end'], rst)

        force_remove(c.stat_path())

        # file_end_handler in iterate()
        rst = []
        force_remove(self.fn)
        append_lines(self.fn, expected)
        for line in fsutil.Cat(self.fn, strip=True, file_end_handler=_end).iterate(timeout=0):
            rst.append(line)
        self.assertEqual(expected + ['end'], rst)

        force_remove(c.stat_path())

        # file_end_handler multi times
        rst = []
        force_remove(self.fn)
        append_lines(self.fn, expected)

        def _append():
            time.sleep(0.1)
            append_lines(self.fn, expected)

        th = threadutil.start_daemon(_append)

        try:
            for line in fsutil.Cat(self.fn, strip=True, file_end_handler=_end).iterate(timeout=0.2):
                rst.append(line)
        except fsutil.NoData:
            pass

        th.join()
        self.assertEqual((expected + ['end']) * 2, rst)

    def test_cat(self):
        self.test_handler()

    def test_strip(self):

        expected = [
            'a' * 32,
            'b' * 32,
        ]
        append_lines(self.fn, expected)

        # by default do not strip
        rst = []
        for line in fsutil.Cat(self.fn).iterate(timeout=0):
            rst.append(line)

        self.assertEqual([x + '\n' for x in expected], rst)

    def test_id(self):
        expected = [
            'a' * 32,
            'b' * 32,
        ]
        append_lines(self.fn, expected)

        force_remove(fsutil.Cat(self.fn, id='foo').stat_path())
        force_remove(fsutil.Cat(self.fn, id='bar').stat_path())

        # different id does not block each other even they cat a same file
        a = fsutil.Cat(self.fn, id='foo', strip=True).iterate(timeout=0)
        b = fsutil.Cat(self.fn, id='bar', strip=True).iterate(timeout=0)

        self.assertEqual(expected[0], a.next())
        self.assertEqual(expected[0], b.next())

    def test_config(self):
        expected = [
            'a' * 32,
            'b' * 32,
        ]
        append_lines(self.fn, expected)

        rc, out, err = proc.shell_script(
            pyt + ' ' + this_base + '/cat/cat_load_config.py ' + self.fn,
            env=dict(PYTHONPATH=this_base + ':' + os.environ.get('PYTHONPATH'))
        )

        dd(rc)
        dd(out)
        dd(err)

        self.assertEqual(0, rc)

        path = [
            'fsutil_cat_lock_',
            'cat_load_config.py',
        ] + os.path.realpath(self.fn)[1:].split(os.path.sep)
        path = './' + '!'.join(path)
        dd(path)

        self.assertTrue(os.path.isfile(path))

        force_remove(path)

    def test_default_seek(self):
        cases = (
            ('abc', {'default_seek': fsutil.SEEK_START}, ['abc']),
            ('abc', {'default_seek': 0},                 ['abc']),
            ('abc', {'default_seek': 1},                 ['bc']),
            ('abc', {'default_seek': -2},                ['c']),
            ('abc', {'default_seek': -100},              ['abc']),
            ('abc', {'default_seek': fsutil.SEEK_END},   []),
            ('abc', {},                                  ['abc']),
        )

        for file_content, kwargs, expected in cases:
            c = fsutil.Cat(self.fn)
            force_remove(c.stat_path())

            append_lines(self.fn, [file_content])

            rst = []
            try:
                kwargs['timeout'] = 0.1
                for l in fsutil.Cat(self.fn, strip=True).iterate(**kwargs):
                    rst.append(l)
            except fsutil.NoData:
                pass

            self.assertEqual(expected, rst)

            force_remove(self.fn)

    def test_max_residual(self):
        c = fsutil.Cat(self.fn)
        force_remove(c.stat_path())

        append_lines(self.fn, ['123'])

        rst = []
        try:
            for l in fsutil.Cat(self.fn, strip=True).iterate(timeout=0.1):
                rst.append(l)
        except fsutil.NoData:
            pass

        self.assertEqual(['123'], rst)

        append_lines(self.fn, ['0' * 1000])

        rst = []
        try:
            for l in fsutil.Cat(self.fn, strip=False).iterate(
                    timeout=0.1, default_seek=-100):
                rst.append(l)
        except fsutil.NoData:
            pass

        self.assertEqual(['0' * 99 + '\n'], rst)

    def test_reset_stat(self):
        cat = fsutil.Cat(self.fn, strip=True)
        force_remove(cat.stat_path())

        append_lines(self.fn, ['123'])

        rst = []
        try:
            for l in cat.iterate(timeout=0.1):
                rst.append(l)
        except fsutil.NoData:
            pass

        self.assertEqual(['123'], rst)

        append_lines(self.fn, ['456'])

        rst = []
        try:
            for l in cat.iterate(timeout=0.1):
                rst.append(l)
        except fsutil.NoData:
            pass

        self.assertEqual(['456'], rst)

        cat.reset_stat()

        rst = []
        try:
            for l in cat.iterate(timeout=0.1):
                rst.append(l)
        except fsutil.NoData:
            pass

        self.assertEqual(['123', '456'], rst)


def force_remove(fn):

    try:
        os.rmdir(fn)
    except:
        pass

    try:
        os.unlink(fn)
    except:
        pass


def append_lines(fn, lines):

    with open(fn, 'a') as f:
        for l in lines:
            f.write(l + '\n')
        f.flush()
        os.fsync(f.fileno())


def append_bytes(fn, *strings):

    with open(fn, 'a') as f:
        for s in strings:
            f.write(s)
        f.flush()
        os.fsync(f.fileno())



#!/usr/bin/env python
# coding: utf-8

import logging

from kazoo.exceptions import NoNodeError

from pykit import rangeset
from pykit import zkutil

from .status import COMMITTED
from .zkaccessor import ZKKeyValue
from .zkaccessor import ZKValue
from .zkstorage import record_nonode_cb
from .zkstorage import txidset_load

logger = logging.getLogger(__name__)


class Slave(object):

    def __init__(self, zke, storage):
        assert isinstance(zke, zkutil.KazooClientExt)

        self.zke = zke
        self.storage = storage
        self.zk_cmt_txidset = None

        self.txidset = ZKValue(self.zke,
                               self.zke._zkconf.txidset,
                               load=txidset_load)

        self.journal = ZKKeyValue(self.zke,
                                  self.zke._zkconf.journal)

        self.record = ZKKeyValue(self.zke,
                                 self.zke._zkconf.record,
                                 nonode_callback=record_nonode_cb)

    def _get_uncommitted_txids(self):
        storage_txidset = self.storage.txidset.get()

        subset = rangeset.substract(self.zk_cmt_txidset, storage_txidset[COMMITTED])

        for begin, end in subset:
            for i in range(begin, end):
                yield i

    def _get_all_records(self):
        record_dir = self.zke._zkconf.record_dir()
        meta_path = ''.join([record_dir, 'meta'])
        names = self.zke.get_children(meta_path)

        rst = {}
        for n in names:
            keys_path = '/'.join([meta_path, n])
            keys = self.zke.get_children(keys_path)

            for k in keys:
                path = '/'.join([keys_path, k])
                val, _ = self.zke.get(path)

                # k='record_dir/meta/server/<server_id>
                # remove record_dir
                path = path[len(record_dir):]
                rst[path] = val

        return rst

    def _set_all_records(self):
        records = self._get_all_records()
        for k, v in records.items():

            while len(v) > 0:
                txid = v[-1][0]
                if not self.zk_cmt_txidset.has(txid):
                    v.pop()
                    logger.warn('txid {txid} not in {cmt_txids}'.format(
                        txid=txid, cmt_txids=self.zk_cmt_txidset))

                else:
                    break

            else:
                logger.error('record {k} txids {txids} not in {cmt_txids}'.format(
                    k=k, txids=v, cmt_txids=self.zk_cmt_txidset))
                continue

            self.storage.apply_record(k, v[-1][1])

        self.storage.set_txidset(COMMITTED, self.zk_cmt_txidset)

    def apply(self):
        sets, _ = self.txidset.get()
        self.zk_cmt_txidset = sets[COMMITTED]

        for txid in self._get_uncommitted_txids():

            try:
                jour, _ = self.journal.get(txid)

            except NoNodeError:
                logger.warn('jour not found txid: {txid}'.format(txid=txid))
                self._set_all_records()
                return

            self.storage.apply_jour(jour)
            self.storage.add_to_txidset(COMMITTED, txid)



from .exceptions import (
    ZKWaitTimeout,
)

from .zkacid import (
    cas_loop,
)

from .zkconf import (
    KazooClientExt,
    ZKConf,

    kazoo_client_ext,
)

from .zkutil import (
    PermTypeError,

    init_hierarchy,
    is_backward_locking,
    lock_id,
    make_acl_entry,
    make_digest,
    make_kazoo_digest_acl,
    parse_kazoo_acl,
    parse_lock_id,
    perm_to_long,
    perm_to_short,
    wait_absent,

    get_next,

)

from .zklock import (
    LockTimeout,

    ZKLock,

)

__all__ = [
    "PermTypeError",
    "ZKWaitTimeout",

    "cas_loop",

    "KazooClientExt",
    "ZKConf",

    "kazoo_client_ext",

    "init_hierarchy",
    "is_backward_locking",
    "lock_id",
    "make_acl_entry",
    "make_digest",
    "make_kazoo_digest_acl",
    "parse_kazoo_acl",
    "parse_lock_id",
    "perm_to_long",
    "perm_to_short",
    "wait_absent",

    "get_next",

    "LockTimeout",

    "ZKLock",
]



#!/usr/bin/env python2
# coding: utf-8

import time
import unittest

from pykit import threadutil
from pykit import ututil

tu = threadutil.threadutil


def _work(sess):
    try:
        while True:
            x = 1

    except SystemExit:
        sess['raised'] = True


class TestThreadutil(unittest.TestCase):

    def _verify_exception_raised(self, sess, thread):
        # we wait few seconds for the exception to be raised.
        for i in range(100):
            if sess['raised']:
                # The thread might still be alive at this point.
                time.sleep(0.2)
                self.assertFalse(thread.is_alive())
                break

            time.sleep(0.1)

        else:
            assert False, "SystemError is not raised"

    def test_raise_in_thread(self):
        sess = {'raised': False}

        t = tu.start_thread(_work, args=(sess, ), daemon=True)
        self.assertFalse(sess['raised'])

        with self.assertRaises(TypeError):
            tu.raise_in_thread('thread', SystemExit)

        with self.assertRaises(TypeError):
            tu.raise_in_thread(t, SystemExit())

        class SomeClass(object):
            pass

        with self.assertRaises(ValueError):
            tu.raise_in_thread(t, SomeClass)

        tu.raise_in_thread(t, SystemExit)
        self._verify_exception_raised(sess, t)

    def test_raise_in_thread_many_times(self):
        sess = {'raised': False}

        t = tu.start_thread(_work, args=(sess, ), daemon=True)
        self.assertFalse(sess['raised'])

        for i in range(5):
            try:
                tu.raise_in_thread(t, SystemExit)
            except tu.InvalidThreadIdError:
                # This will happen if the thread is already terminated by
                # a previous raise_in_thread call.
                pass

        self._verify_exception_raised(sess, t)

        # Raising in a dead thread shoud not break.
        with self.assertRaises(tu.InvalidThreadIdError):
            tu.raise_in_thread(t, SystemExit)

    def test_start_thread(self):
        def _sort(a, reverse=False):
            a.sort(reverse=reverse)

        array = [3, 1, 2]
        t = tu.start_thread(_sort, args=(array, ))
        t.join()

        self.assertEqual(array, [1, 2, 3])

        t = tu.start_thread(_sort, args=(array, ),
                            kwargs={'reverse': True})
        t.join()

        self.assertEqual(array, [3, 2, 1])

    def test_start_daemon(self):
        def noop(): return None

        # Thread should be non-daemon by default
        t = tu.start_thread(noop)
        self.assertFalse(t.daemon)

        t = tu.start_thread(noop, daemon=True)
        self.assertTrue(t.daemon)

        t = tu.start_daemon(noop)
        self.assertTrue(t.daemon)

    def test_thread_after(self):

        def _do():
            pass

        with ututil.Timer() as t:
            th = tu.start_thread(target=_do, after=None)
            th.join()
            self.assertAlmostEqual(0, t.spent(), delta=0.1)

        with ututil.Timer() as t:
            th = tu.start_thread(target=_do, after=0.5)
            th.join()
            self.assertAlmostEqual(0.5, t.spent(), delta=0.1)

    def test_daemon_after(self):

        def _do():
            pass

        with ututil.Timer() as t:
            th = tu.start_daemon(target=_do, after=None)
            th.join()
            self.assertAlmostEqual(0, t.spent(), delta=0.1)

        with ututil.Timer() as t:
            th = tu.start_daemon(target=_do, after=0.5)
            th.join()
            self.assertAlmostEqual(0.5, t.spent(), delta=0.1)



#!/usr/bin/env python2
# coding: utf-8

import unittest

from pykit import cluster


class TestClusterBlock(unittest.TestCase):

    def test_parse_and_print(self):

        block_id = 'd1g0006300000001230101c62d8736c72800020000000001'

        bid = cluster.BlockID('d1', 'g000630000000123', '0101', cluster.DriveID.parse('c62d8736c7280002'), 1)
        self.assertEqual(block_id, str(bid))

        bid = cluster.BlockID.parse(block_id)

        self.assertEqual('d1', bid.type)
        self.assertEqual('g000630000000123', str(bid.block_group_id))
        self.assertEqual((1, 1), bid.block_index)
        self.assertEqual('0101', str(bid.block_index))
        self.assertEqual('c62d8736c7280002', bid.drive_id.tostr())
        self.assertEqual(1, bid.bg_seq)

        self.assertEqual(block_id, str(bid))
        self.assertEqual(block_id, '{0}'.format(bid))
        self.assertEqual(
            "_BlockID(type='d1', block_group_id=_BlockGroupID(block_size=63, seq=123), block_index=_BlockIndex(i=1, j=1), drive_id=_DriveID(server_id='c62d8736c728', mountpoint_index=2), bg_seq=1)",
            repr(bid))

        # test invalid input
        block_id_invalid = 'd1g0006300000001230101c62d8736c728000200000'
        self.assertRaises(cluster.BlockIDError,
                          cluster.BlockID.parse, block_id_invalid)

    def test_new(self):

        block_id = 'd1g0006300000001230101c62d8736c72800020000000001'

        # new with string drive id

        bid = cluster.BlockID('d1', 'g000630000000123', '0101', 'c62d8736c7280002', 1)
        self.assertEqual(block_id, str(bid))

        self.assertIsInstance(bid.drive_id, cluster.DriveID)

    def test_tostr(self):

        block_id = 'd1g0006300000001230101c62d8736c72800020000000001'
        bid = cluster.BlockID.parse(block_id)

        self.assertEqual(block_id, bid.tostr())
        self.assertEqual(block_id, str(bid))

        self.assertIsInstance(bid.drive_id, cluster.DriveID)
        self.assertEqual('c62d8736c7280002', str(bid.drive_id))
        self.assertEqual('c62d8736c7280002', bid.drive_id.tostr())
        self.assertEqual('c62d8736c728', bid.drive_id.server_id)



#!/usr/bin/env python2.6
# coding: utf-8

import unittest

from pykit import awssign


class TestAddAuth(unittest.TestCase):

    def test_basic(self):
        signer = awssign.Signer('access_key', 'secret_key')
        request = {
            'verb': 'GET',
            'uri': '/',
            'args': {
                'foo': 'bar',
                'acl': True,
            },
            'headers': {
                'host': '127.0.0.1',
            },
            'body': 'foo',
        }

        signer.add_auth(request, sign_payload=True,
                        request_date='20180101T120101Z')
        self.assertEqual('/?acl&foo=bar', request['uri'])
        self.assertEqual('AWS4-HMAC-SHA256 Credential=access_key/20180101/us-east-1/s3/aws4_request, SignedHeaders=host;x-amz-content-sha256;x-amz-date, Signature=c0b1d89ddd41df96454d3a5e2c82afdc44aa19bc6593d4fa54bc277756dcc3ef',
                         request['headers']['Authorization'])
        self.assertEqual('2c26b46b68ffc68ff99b453c1d30413413422d706483bfa0f98a5e886266e7ae',
                         request['headers']['X-Amz-Content-SHA256'])
        self.assertEqual('20180101T120101Z', request['headers']['X-Amz-Date'])

    def test_query_auth(self):
        signer = awssign.Signer('access_key', 'secret_key')
        request = {
            'verb': 'GET',
            'uri': '/',
            'args': {
                'foo': 'bar',
                'acl': True,
            },
            'headers': {
                'host': '127.0.0.1',
            },
            'body': 'foo',
        }

        signer.add_auth(request, sign_payload=True, query_auth=True,
                        request_date='20180101T120101Z')
        self.assertEqual('/?acl&foo=bar&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Expires=60&X-Amz-Credential=access_key%2F20180101%2Fus-east-1%2Fs3%2Faws4_request&X-Amz-SignedHeaders=host&X-Amz-Date=20180101T120101Z&X-Amz-Signature=7aaac5758f74924ef89431dccebebf65a01e9a932b1eb875422c05d7b46efd86',
                         request['uri'])

    def test_unicode(self):
        unicode_str = '测试'.decode('utf-8')
        signer = awssign.Signer(unicode_str, unicode_str, region=unicode_str,
                                service=u's3')
        request = {
            'verb': u'GET',
            'uri': '/' + unicode_str,
            'args': {
                unicode_str: unicode_str,
            },
            'headers': {
                'Host': '127.0.0.1',
                'x-amz-content-sha256': unicode_str,
                unicode_str: unicode_str,
                u'foo': u'bar',
            },
            'body': unicode_str,
        }

        ctx = signer.add_auth(request, headers_not_to_sign=[unicode_str],
                              request_date=u'20190101T120000Z',
                              signing_date=u'20180101',
                              sign_payload=True)

        self.assertEqual(unicode_str.encode('utf-8'),
                         request['headers']['X-Amz-Content-SHA256'])
        self.assertIsInstance(request['headers']['Authorization'], str)
        self.assertIsInstance(request['headers']['X-Amz-Date'], str)
        self.assertEqual('20190101T120000Z', ctx['request_date'])
        self.assertEqual(
            'foo;host;x-amz-content-sha256;x-amz-date', ctx['signed_headers'])



#!/usr/bin/env python
# coding: utf-8

import logging

from kazoo.exceptions import NodeExistsError
from kazoo.exceptions import NoNodeError

logger = logging.getLogger(__name__)


class ZKKeyValue(object):
    def __init__(self, zkclient, get_path=None, load=None, dump=None, nonode_callback=None):

        self.zkclient = zkclient
        self._get_path = get_path
        self.load = load
        self.dump = dump
        self.nonode_callback = nonode_callback

    def get_path(self, key):
        if self._get_path is None:
            return key
        else:
            return self._get_path(key)

    def create(self, key, value, ephemeral=False, sequence=False):

        value = self._dump(value)
        return self.zkclient.create(self.get_path(key), value, acl=self._get_acl(),
                                    ephemeral=ephemeral, sequence=sequence)

    def delete(self, key, version=-1):
        return self.zkclient.delete(self.get_path(key), version=version)

    def set(self, key, value, version=-1):
        value = self._dump(value)
        self.zkclient.set(self.get_path(key), value, version=version)

    def set_or_create(self, key, value, version=-1):

        value = self._dump(value)
        while True:
            try:
                self.zkclient.set(self.get_path(key), value, version=version)
                return
            except NoNodeError:
                if version == -1:
                    try:
                        self.zkclient.create(self.get_path(key), value, acl=self._get_acl())
                        return
                    except NodeExistsError:
                        continue
                else:
                    raise

    def get(self, key):
        try:
            val, zstat = self.zkclient.get(self.get_path(key))
            version = zstat.version
        except NoNodeError:
            if self.nonode_callback is not None:
                val, version = self.nonode_callback()
            else:
                raise
        return self._load(val), version

    def _load(self, val):
        if self.load is not None:
            return self.load(val)
        else:
            return val

    def _dump(self, val):
        if self.dump is not None:
            return self.dump(val)
        else:
            return val

    def _get_acl(self):
        acl = None
        zkconf = getattr(self.zkclient, '_zkconf', None)
        if zkconf is not None:
            acl = zkconf.kazoo_digest_acl()
        return acl


class ZKValue(ZKKeyValue):
    def get_path(self, key):
        return self._get_path()

    def create(self, value, ephemeral=False, sequence=False):
        return super(ZKValue, self).create('', value,
                                           ephemeral=ephemeral, sequence=sequence)

    def delete(self, version=-1):
        return super(ZKValue, self).delete('', version=version)

    def set(self, value, version=-1):
        return super(ZKValue, self).set('', value, version=version)

    def set_or_create(self, value, version=-1):
        return super(ZKValue, self).set_or_create('', value, version=version)

    def get(self):
        return super(ZKValue, self).get('')



from .net import (

    INN,
    PUB,

    LOCALHOST,

    NetworkError,
    IPUnreachable,
    InvalidIP4,
    InvalidIP4Number,

    choose_by_idc,
    choose_inn,
    choose_pub,
    choose_by_regex,
    get_host_devices,
    get_host_ip4,
    ip_class,
    ips_prefer,
    is_inn,
    is_pub,
    is_ip4,
    is_ip4_loopback,
    parse_ip_regex_str,
    ip_to_num,
    num_to_ip,
)

__all__ = [
    'INN',
    'PUB',

    'LOCALHOST',

    'NetworkError',
    'IPUnreachable',
    'InvalidIP4',
    'InvalidIP4Number',

    'choose_by_idc',
    'choose_inn',
    'choose_pub',
    'choose_by_regex',
    'get_host_devices',
    'get_host_ip4',
    'ip_class',
    'ips_prefer',
    'is_inn',
    'is_pub',
    'is_ip4',
    'is_ip4_loopback',
    'parse_ip_regex_str',
    'ip_to_num',
    'num_to_ip',
]



import logging

logger = logging.getLogger(__name__)


def run(context):
    queue = context['queue']

    while True:
        log_entry = queue.get()

        try:
            context['send_log'](log_entry)
        except Exception:
            # do not log repr(e), it may cause endless escape of '\'.
            logger.error('failed to send log entry')



import logging
import os

from pykit import fsutil

logger = logging.getLogger(__name__)


def create_cgroup(cgroup_path):
    os.mkdir(cgroup_path, 0755)
    return


def add_pids(cgroup_path, pids):
    task_file = os.path.join(cgroup_path, 'tasks')

    for pid in pids:
        try:
            fsutil.write_file(task_file, str(pid), fsync=False)
        except Exception as e:
            logger.info('failed to add pid: %s to file: %s, %s' %
                        (str(pid), task_file, repr(e)))
    return


def clear_pids(subsystem_dir, cgroup_path):
    task_file = os.path.join(cgroup_path, 'tasks')
    f = open(task_file, 'r')

    while True:
        line = f.readline()
        if line == '':
            break

        pid = line.strip()

        add_pids(subsystem_dir, [pid])

    return


def remove_cgroup(subsystem_dir, cgroup_path):
    sub_dirs = fsutil.get_sub_dirs(cgroup_path)

    for sub_dir in sub_dirs:
        sub_cgroup_path = os.path.join(cgroup_path, sub_dir)
        remove_cgroup(subsystem_dir, sub_cgroup_path)

    clear_pids(subsystem_dir, cgroup_path)
    os.rmdir(cgroup_path)

    return



import errno
import logging
import logging.handlers
import os
import sys
import traceback
from stat import ST_DEV
from stat import ST_INO

import __main__
from pykit import config

logger = logging.getLogger(__name__)

log_formats = {
    'default': '[%(asctime)s,%(process)d-%(thread)d,%(filename)s,%(lineno)d,%(levelname)s] %(message)s',
    'time_level': "[%(asctime)s,%(levelname)s] %(message)s",
    'message': '%(message)s',

    # more info: but it is too long.
    # 'full': '[%(asctime)s,%(process)d-%(thread)d,%(name)s, %(filename)s,%(lineno)d,%(funcName)s %(levelname)s] %(message)s',
}

date_formats = {
    # by default do not specify
    'default': None,
    'time': '%H:%M:%S',
}

log_suffix = 'out'

default_stack_sep = ' --- '


class FixedWatchedFileHandler(logging.FileHandler):
    """
    A handler for logging to a file, which watches the file
    to see if it has changed while in use. This can happen because of
    usage of programs such as newsyslog and logrotate which perform
    log file rotation. This handler, intended for use under Unix,
    watches the file to see if it has changed since the last emit.
    (A file has changed if its device or inode have changed.)
    If it has changed, the old file stream is closed, and the file
    opened to get a new stream.

    This handler is not appropriate for use under Windows, because
    under Windows open files cannot be moved or renamed - logging
    opens the files with exclusive locks - and so there is no need
    for such a handler. Furthermore, ST_INO is not supported under
    Windows; stat always returns zero for this value.

    This handler is based on a suggestion and patch by Chad J.
    Schroeder.
    """

    def __init__(self, filename, mode='a', encoding=None, delay=0):
        logging.FileHandler.__init__(self, filename, mode, encoding, delay)
        self.dev, self.ino = -1, -1
        self._statstream()

    def _statstream(self):
        if self.stream:
            sres = os.fstat(self.stream.fileno())
            self.dev, self.ino = sres[ST_DEV], sres[ST_INO]

    def emit(self, record):
        """
        Emit a record.

        First check if the underlying file has changed, and if it
        has, close the old stream and reopen the file to get the
        current stream.
        """
        # Reduce the chance of race conditions by stat'ing by path only
        # once and then fstat'ing our new fd if we opened a new log stream.
        # See issue #14632: Thanks to John Mulligan for the problem report
        # and patch.
        try:
            # stat the file by path, checking for existence
            sres = os.stat(self.baseFilename)
        except OSError as e:
            if e.errno == errno.ENOENT:
                sres = None
            else:
                raise
        # compare file system stat with that of our stream file handle
        if not sres or sres[ST_DEV] != self.dev or sres[ST_INO] != self.ino:

            # Fixed by xp 2017 Apr 03:
            #     os.fstat still gets OSError(errno=2), although it operates
            #     directly on fd instead of path.  The same for stream.flush().
            #     Thus we keep on trying this close/open/stat loop until no
            #     OSError raises.

            for ii in range(16):
                try:
                    if self.stream is not None:
                        # we have an open file handle, clean it up
                        self.stream.flush()
                        self.stream.close()
                        # See Issue #21742: _open () might fail.
                        self.stream = None
                        # open a new file handle and get new stat info from
                        # that fd
                        self.stream = self._open()
                        self._statstream()
                        break
                except OSError as e:
                    if e.errno == errno.ENOENT:
                        continue
                    else:
                        raise
        logging.FileHandler.emit(self, record)


def get_root_log_fn():

    if hasattr(__main__, '__file__'):
        name = __main__.__file__
        name = os.path.basename(name)
        if name == '<stdin>':
            name = '__stdin__'
        return name.rsplit('.', 1)[0] + '.' + log_suffix
    else:
        return '__instant_command__.' + log_suffix


def make_logger(base_dir=None, log_name=None, log_fn=None,
                level=logging.DEBUG, fmt=None,
                datefmt=None):

    # if log_name is None, get the root logger
    logger = logging.getLogger(log_name)
    logger.setLevel(level)

    if base_dir is None:
        base_dir = config.log_dir

    # do not add 2 handlers to one logger by default
    if len(logger.handlers) == 0:

        if log_fn is None:
            if log_name is None:
                log_fn = get_root_log_fn()
            else:
                log_fn = log_name + '.' + log_suffix

        logger.addHandler(make_file_handler(base_dir, log_fn,
                                            fmt=fmt, datefmt=datefmt))

    return logger


def make_file_handler(base_dir=None, log_fn=None, fmt=None, datefmt=None):

    if base_dir is None:
        base_dir = config.log_dir
    if log_fn is None:
        log_fn = get_root_log_fn()
    file_path = os.path.join(base_dir, log_fn)

    handler = FixedWatchedFileHandler(file_path)
    handler.setFormatter(make_formatter(fmt=fmt, datefmt=datefmt))

    return handler


def set_logger_level(level=logging.INFO, name_prefixes=None):
    if name_prefixes is None:
        name_prefixes = ('',)

    root_logger = logging.getLogger()

    loggers = root_logger.manager.loggerDict.items()
    loggers.sort()

    for name, _ in loggers:
        if name.startswith(name_prefixes):
            name_logger = logging.getLogger(name)
            name_logger.setLevel(level)


def add_std_handler(logger, stream=None, fmt=None, datefmt=None, level=None):

    stream = stream or sys.stdout

    if stream == 'stdout':
        stream = sys.stdout

    elif stream == 'stderr':
        stream = sys.stderr

    stdhandler = logging.StreamHandler(stream)
    stdhandler.setFormatter(make_formatter(fmt=fmt, datefmt=datefmt))
    if level is not None:
        stdhandler.setLevel(level)

    logger.addHandler(stdhandler)

    return logger


def make_formatter(fmt=None, datefmt=None):

    fmt = get_fmt(fmt)
    datefmt = get_datefmt(datefmt)

    return logging.Formatter(fmt=fmt, datefmt=datefmt)


def get_fmt(fmt):

    if fmt is None:
        fmt = 'default'

    return log_formats.get(fmt, fmt)


def get_datefmt(datefmt):

    if datefmt is None:
        return datefmt

    return date_formats.get(datefmt, datefmt)


def stack_list(offset=0):
    offset += 1  # count this function as 1

    # list of ( filename, line-nr, in-what-function, statement )
    x = traceback.extract_stack()
    return x[: -offset]


def stack_format(stacks, fmt=None, sep=None):

    if fmt is None:
        fmt = "{fn}:{ln} in {func} {statement}"

    if sep is None:
        sep = default_stack_sep

    dict_stacks = []
    for st in stacks:
        o = {
            'fn': os.path.basename(st[0]),
            'ln': st[1],
            'func': st[2],
            'statement': st[3],
        }
        dict_stacks.append(o)

    return sep.join([fmt.format(**xx)
                     for xx in dict_stacks])


def stack_str(offset=0, fmt=None, sep=None):
    offset += 1  # count this function as 1
    return stack_format(stack_list(offset), fmt=fmt, sep=sep)


def deprecate(msg=None, fmt=None, sep=None):

    d = 'Deprecated:'

    if msg is not None:
        d += ' ' + str(msg)

    logger.warn(d + (sep or default_stack_sep)
                + stack_str(offset=1, fmt=fmt, sep=sep))



#!/usr/bin/env python
# coding: utf-8

import logging
import threading
import time

from kazoo.client import KazooClient
from kazoo.exceptions import KazooException
from kazoo.exceptions import LockTimeout
from kazoo.exceptions import NodeExistsError
from kazoo.exceptions import NoNodeError

from . import zkutil
from .zkconf import ZKConf

logger = logging.getLogger(__name__)


class ZKTXError(Exception):
    pass


class ZKTXConnectionLost(ZKTXError):
    pass


# TODO add method: ZKLock.get_owner() to get current lock owner
class ZKLock(object):

    def __init__(self, lock_name,
                 zkconf=None,
                 zkclient=None,
                 on_lost=None,
                 identifier=None,
                 ephemeral=True,
                 timeout=10):

        if zkconf is None:
            zkconf = ZKConf()
        if isinstance(zkconf, dict):
            zkconf = ZKConf(**zkconf)
        self.zkconf = zkconf

        if zkclient is None:
            # If user does not pass a zkclient instance,
            # we need to create one for lock to use.
            # This zkclient will be closed after lock is released

            if on_lost is None:
                raise ValueError('on_lost must be specified to watch zk connection issue if no zkclient specified')

            zkclient = make_owning_zkclient(zkconf.hosts(), zkconf.auth())
            self.owning_client = True
        else:
            self.owning_client = False

        if identifier is None:
            identifier = zkutil.lock_id(zkconf.node_id())

        # a copy of hosts for debugging and tracking
        self._hosts = ','.join(['{0}:{1}'.format(*x)
                                for x in zkclient.hosts])

        self.zkclient = zkclient
        self.on_lost = on_lost

        self.lock_name = lock_name
        self.lock_path = zkconf.lock(self.lock_name)
        self.identifier = identifier
        self.ephemeral = ephemeral
        self.timeout = timeout

        self.mutex = threading.RLock()
        self.maybe_available = threading.Event()
        self.maybe_available.set()
        self.lock_holder = None

        logger.info('adding event listener: {s}'.format(s=self))
        self.zkclient.add_listener(self.on_connection_change)

    def on_node_change(self, watchevent):

        # Must be locked first.
        # Or there is a chance on_node_change is triggered before
        #         self.maybe_available.clear()
        with self.mutex:
            self.maybe_available.set()

            # If locked. the node change is treated as losing a lock
            if self.is_locked():
                if self.on_lost is not None:
                    self.on_lost()

        logger.info('node state changed:{ev}, lock might be released: {s}'.format(ev=watchevent, s=str(self)))

    def on_connection_change(self, state):

        # notify zklock to re-do acquiring procedure, to trigger Connection Error
        with self.mutex:
            self.maybe_available.set()

        if self.on_lost is not None:
            self.on_lost()

    def acquire_loop(self, timeout=None):

        if timeout is None:
            timeout = self.timeout

        expire_at = time.time() + timeout

        while True:

            # Even if timeout is smaller than 0, try-loop continue on until
            # maybe_available is not ready.
            #
            # There is a chance that:
            #  - Failed to create lock node(lock is occupied by other)
            #  - Failed to get lock node(just deleted)
            #  - Failed to create lock node(lock is occupied by other)
            #  - Failed to get lock node(just deleted)
            #  - ...
            if not self.maybe_available.wait(timeout=expire_at - time.time()):

                logger.debug('lock is still held by others: ' + str(self))

                if time.time() > expire_at:
                    raise LockTimeout('lock: ' + str(self.lock_path))

            # Always proceed the "get" phase, in order to add a watch handler.
            # To watch node change event.

            self._create()
            self._acquire_by_get()
            if self.is_locked():
                return

            # If it is possible to acquire the lock in next retry, do not yield
            if self.maybe_available.is_set():
                continue
            else:
                yield self.lock_holder[0], self.lock_holder[1]

    def acquire(self, timeout=None):

        for holder, ver in self.acquire_loop(timeout=timeout):
            continue

    def try_acquire(self):

        try:
            self.acquire(timeout=-1)
        except LockTimeout:
            pass

        # if_locked, lock holder identifier, holder version
        return self.is_locked(), self.lock_holder[0], self.lock_holder[1]

    def try_release(self):

        logger.debug('try to release if I am locker holder: {s}'.format(s=str(self)))

        try:
            holder, zstat = self.zkclient.get(self.lock_path)
            self.lock_holder = (holder, zstat.version)

            logger.debug('got lock holder: {s}'.format(s=str(self)))

            if holder == self.identifier:

                self.zkclient.remove_listener(self.on_connection_change)

                try:
                    self.zkclient.delete(self.lock_path, version=zstat.version)
                except NoNodeError as e:
                    logger.info(repr(e) + ' while delete lock: ' + str(self))

                self.lock_holder = None

                return True, holder, zstat.version
            else:
                return False, holder, zstat.version

        except NoNodeError as e:
            logger.info(repr(e) + ' while try_release: {s}'.format(s=str(self)))
            return True, self.identifier, -1

    def release(self):

        with self.mutex:

            if self.is_locked():

                # remove listener to avoid useless event triggering
                self.zkclient.remove_listener(self.on_connection_change)

                try:
                    self.zkclient.delete(self.lock_path)
                except NoNodeError as e:
                    logger.info(repr(e) + ' while delete lock: ' + str(self))

                self.lock_holder = None

                logger.info('RELEASED: {s}'.format(s=str(self)))
            else:
                logger.info('not acquired, do not need to release')

        self.close()

    def close(self):

        self.zkclient.remove_listener(self.on_connection_change)

        if self.owning_client:

            logger.info('zk client is made by me, close it')

            try:
                self.zkclient.stop()
            except KazooException as e:
                logger.info(repr(e) + ' while stop my own client')

    def is_locked(self):

        l = self.lock_holder

        return (l is not None
                and l[0] == self.identifier)

    def _create(self):

        logger.debug('to creaet: {s}'.format(s=str(self)))

        try:
            self.zkclient.create(self.lock_path,
                                 self.identifier,
                                 ephemeral=self.ephemeral,
                                 acl=self.zkconf.kazoo_digest_acl())

        except NodeExistsError as e:

            # NOTE Success create on server side might also results in failure
            # on client side due to network issue.
            # 'get' after 'create' to check if existent node belongs to this
            # client.

            logger.debug(repr(e) + ' while create lock: {s}'.format(s=str(self)))
            self.lock_holder = None
            return

        logger.info('CREATE OK: {s}'.format(s=str(self)))

    def _acquire_by_get(self):

        logger.debug('to get: {s}'.format(s=str(self)))

        try:
            with self.mutex:

                holder, zstat = self.zkclient.get(self.lock_path, watch=self.on_node_change)
                self.lock_holder = (holder, zstat.version)

                logger.debug('got lock holder: {s}'.format(s=str(self)))

                if holder == self.identifier:
                    logger.info('ACQUIRED: {s}'.format(s=str(self)))
                    return

                logger.debug('other holds: {s}'.format(s=str(self)))
                self.maybe_available.clear()

        except NoNodeError as e:
            # create failed but when getting  it, it has been deleted
            logger.info(repr(e) + ' while get lock: {s}'.format(s=str(self)))
            with self.mutex:
                self.lock_holder = None
                self.maybe_available.set()

    def __str__(self):
        return '<id={id} {l}:[{holder}] on {h}>'.format(
            id=self.identifier,
            l=self.lock_path,
            holder=(self.lock_holder or ''),
            h=str(self._hosts),
        )

    def __enter__(self):
        self.acquire()

    def __exit__(self, tp, value, tb):
        self.release()


def make_owning_zkclient(hosts, auth):

    zkclient = KazooClient(hosts=hosts)
    zkclient.start()

    if auth is not None:
        scheme, name, passw = auth
        zkclient.add_auth(scheme, name + ':' + passw)

    return zkclient



import mock
import os
import sys
import time
import unittest
import urlparse

from BaseHTTPServer import BaseHTTPRequestHandler
from BaseHTTPServer import HTTPServer

from pykit import redisutil
from pykit import threadutil
from pykit import utdocker
from pykit import ututil
from pykit import utfjson

dd = ututil.dd

redis_tag = 'daocloud.io/redis:3.2.3'

redis_port = 6379


class TestRedis(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        utdocker.pull_image(redis_tag)

    def setUp(self):

        self.containers = [
            ('redis-0', redis_tag, '192.168.52.40'),
            ('redis-1', redis_tag, '192.168.52.41'),
        ]

        # for single redis cases:
        self.docker_name = 'redis-0'
        self.ip = '192.168.52.40'
        self.is_child = False

        utdocker.create_network()

        for args in self.containers:
            utdocker.start_container(*(args + ('',)))
            dd('started redis in docker: ' + repr(args))

            redisutil.wait_serve((args[2], redis_port))

        self.rcl = redisutil.get_client((self.ip, redis_port))

    def tearDown(self):

        # do not tear down in child process
        if self.is_child:
            return

        for args in self.containers:
            dd('remove_container: ' + args[0])
            utdocker.remove_container(args[0])

    def test_set_get(self):

        hname = 'foo'

        rst = self.rcl.hset(hname, 'a', '1')
        dd('hset rst:', rst)

        rst = self.rcl.hget(hname, 'a')
        dd('hget rst:', rst)

        self.assertEqual('1', rst)

    def test_recreate_client(self):

        hname = 'foo'

        pid = os.fork()

        dd('my pid:', os.getpid())

        if pid == 0:
            # child
            self.is_child = True
            rcl = redisutil.get_client((self.ip, redis_port))
            rcl.hset(hname, 'a', '5')

            child_exit()

        else:
            # parent

            os.waitpid(pid, 0)
            dd('child pid quit: ' + repr(pid))

            rst = self.rcl.hget(hname, 'a')

            dd('redis hget rst:', rst)

            self.assertEqual('5', rst)

    def test_separated_connection_after_fork(self):

        hname = 'foo'

        pid = os.fork()
        dd('my pid:', os.getpid())
        n = 10240

        if pid == 0:
            # child
            self.rcl = redisutil.get_client((self.ip, redis_port))

        # for both parent and child process
        for i in range(n):
            self.rcl.hset(hname, 'x' + str(i),
                          'foobarjfkdslafjdasklfdjsaklfdsa' + str(i))
            self.rcl.hget(hname, 'x' + str(i - 1))

            if i % 1024 == 0:
                dd('pid:', os.getpid(), ' finished ', i, ' set/get')

        if pid == 0:
            # child
            self.is_child = True
            child_exit()
        else:
            # parent
            os.waitpid(pid, 0)
            dd('child pid quit: ' + repr(pid))

        # no error raised with 2 process hget/hset, it is ok

    def test_2_ip_port(self):
        rcl = []
        for name, tag, ip in self.containers:
            rcl.append(redisutil.get_client((ip, redis_port)))

        rcl[0].hset('foo', 'foo', '100')
        rcl[1].hset('foo', 'foo', '101')

        rst = rcl[0].hget('foo', 'foo')
        dd('rst got from first redis:', rst)

        self.assertEqual('100', rst)

    def test_normalize_ip_port(self):

        self.assertEqual(('127.0.0.1', 1234),
                         redisutil.normalize_ip_port(1234))
        self.assertEqual(('1.2.3.4', 1234),
                         redisutil.normalize_ip_port(('1.2.3.4', 1234)))

    def test_redis_channel(self):

        c = redisutil.RedisChannel((self.ip, redis_port), '/foo', 'client')
        s = redisutil.RedisChannel((self.ip, redis_port), '/foo', 'server')

        rst = s.peek_msg()
        dd('server peek: ', rst)
        self.assertEqual(None, rst)

        c.send_msg('c2s')
        s.send_msg('s2c')

        # peek does not remove message

        for ii in range(2):
            rst = s.peek_msg()
            dd('server peek: ', rst)
            self.assertEqual('c2s', rst)

            rst = c.peek_msg()
            dd('client peek: ', rst)
            self.assertEqual('s2c', rst)

        rst = s.recv_msg()
        dd('server recv: ', rst)
        self.assertEqual('c2s', rst, 'server pop message from client')

        rst = s.peek_msg()
        dd('server peek: ', rst)
        self.assertEqual(None, rst, 'no more message in channel')

        # recv_last_msg

        for ii in range(10):
            s.send_msg('s2c' + str(ii))

        rst = c.recv_last_msg()
        dd('client recv last:', rst)
        self.assertEqual('s2c9', rst)

        rst = c.peek_msg()
        self.assertEqual(
            None, rst, 'all messages should have been read, after recv_last_msg()')

    def test_list_redis_channel(self):

        ca = redisutil.RedisChannel((self.ip, redis_port), '/foo/a', 'client')
        sb = redisutil.RedisChannel((self.ip, redis_port), '/foo/b', 'server')

        rst = ca.list_channel('/foo/')
        dd('before write, list_channel:', rst)
        self.assertEqual([], rst, 'can not list list before any message sent')

        ca.send_msg(1)
        sb.send_msg(2)

        rst = ca.list_channel('/foo/')
        dd('after write, list_channel:', rst)
        self.assertEqual(['/foo/a', '/foo/b'], rst)

        # empty channel then channel can not be seen.
        redisutil.RedisChannel((self.ip, redis_port),
                               '/foo/a', 'server').recv_msg()
        redisutil.RedisChannel((self.ip, redis_port),
                               '/foo/b', 'client').recv_msg()

        rst = ca.list_channel('/foo/')
        self.assertEqual([], rst)

    def test_tuple_channel(self):

        c = redisutil.RedisChannel((self.ip, redis_port), '/foo/a', 'client')
        s = redisutil.RedisChannel(
            (self.ip, redis_port), ('foo', 'a'), 'server')

        c.send_msg(1)
        rst = s.recv_msg()
        dd('server recv:', rst)

        self.assertEqual(1, rst)

    def test_channel_timeout(self):

        ca = redisutil.RedisChannel((self.ip, redis_port), '/foo/a', 'client', timeout=1)
        ca.send_msg(1)
        self.assertEqual(['/foo/a'], ca.list_channel('/foo/'))
        time.sleep(1.5)
        self.assertEqual([], ca.list_channel('/foo/'))

    def test_brecv_message(self):

        c = redisutil.RedisChannel((self.ip, redis_port), '/foo', 'client')
        s = redisutil.RedisChannel((self.ip, redis_port), '/foo', 'server')

        c.send_msg('aa')
        self.assertEqual('aa', s.brecv_msg(timeout=1))
        self.assertEqual(None, s.brecv_msg(timeout=1))

        def _send_msg():
            time.sleep(0.5)
            c.send_msg('bar')

        threadutil.start_daemon_thread(target=_send_msg)
        self.assertEqual('bar', s.brecv_msg(timeout=1))

    def test_brecv_last_message(self):

        c = redisutil.RedisChannel((self.ip, redis_port), '/foo', 'client')
        s = redisutil.RedisChannel((self.ip, redis_port), '/foo', 'server')

        c.send_msg('aa')
        c.send_msg('bb')
        self.assertEqual('bb', s.brecv_last_msg(timeout=1))
        self.assertEqual(None, s.brecv_last_msg(timeout=1))

        def _send_msg():
            time.sleep(0.5)
            c.send_msg('cc')

        threadutil.start_daemon_thread(target=_send_msg)
        self.assertEqual('cc', s.brecv_last_msg(timeout=1))

    def test_rpeek_message(self):

        c = redisutil.RedisChannel((self.ip, redis_port), '/foo', 'client')
        s = redisutil.RedisChannel((self.ip, redis_port), '/foo', 'server')

        self.assertEqual(None, c.rpeek_msg())
        self.assertEqual(None, s.rpeek_msg())

        c.send_msg('c2s1')
        c.send_msg('c2s2')
        s.send_msg('s2c1')
        s.send_msg('s2c2')

        # rpeek does not remove message
        for ii in range(2):
            self.assertEqual('c2s2', s.rpeek_msg())
            self.assertEqual('s2c2', c.rpeek_msg())


class TestRedisProxyClient(unittest.TestCase):

    response = {}
    request = {}
    access_key = 'test_accesskey'
    secret_key = 'test_secretkey'

    def setUp(self):
        self.addr = ('127.0.0.1', 22038)

        self.response['http-status'] = 200

        def _start_http_svr():
            self.http_server = HTTPServer(self.addr, HttpHandle)
            self.http_server.serve_forever()

        threadutil.start_daemon_thread(_start_http_svr)
        time.sleep(0.1)

        self.cli = redisutil.RedisProxyClient([self.addr])
        self.n = self.cli.n
        self.w = self.cli.w
        self.r = self.cli.r

    def tearDown(self):
        self.http_server.shutdown()
        self.http_server.server_close()

    def test_send_request_failed(self):
        # close http server
        self.tearDown()

        self.assertRaises(redisutil.SendRequestError, self.cli.get, 'foo')
        self.assertRaises(redisutil.SendRequestError, self.cli.set, 'foo', 'bar')

        self.assertRaises(redisutil.SendRequestError, self.cli.hget, 'foo', 'bar')
        self.assertRaises(redisutil.SendRequestError, self.cli.hset, 'foo', 'bar', 'xx')

    def test_not_found(self):
        self.response['http-status'] = 404
        self.assertRaises(redisutil.KeyNotFoundError, self.cli.get, 'foo')
        self.assertRaises(redisutil.KeyNotFoundError, self.cli.hget, 'foo', 'bar')

    def test_server_response_error(self):
        cases = (
            201,
            302,
            403,
            500,
        )

        for status in cases:
            self.response['http-status'] = status
            self.assertRaises(redisutil.ServerResponseError, self.cli.get, 'foo')
            self.assertRaises(redisutil.ServerResponseError, self.cli.hget, 'foo', 'bar')
            self.assertRaises(redisutil.ServerResponseError, self.cli.set, 'foo', 'val')
            self.assertRaises(redisutil.ServerResponseError, self.cli.hset, 'foo', 'bar', 'val')

    def test_get(self):
        cases = (
            ('foo', None),
            ('bar', 1),
            ('123', 2),
            ('foobar123', 3)
        )

        for key, retry_cnt in cases:
            res = self.cli.get(key, retry_cnt)
            time.sleep(0.1)

            exp_path = '{ver}/GET/{k}'.format(ver=self.cli.ver, k=key)
            self.assertEqual(exp_path, TestRedisProxyClient.request['req-path'])

            exp_qs = 'n=3&w=2&r=2'
            self.assertIn(exp_qs, TestRedisProxyClient.request['req-qs'])

            exp_res = {
                'foo': 1,
                'bar': 2,
            }
            self.assertDictEqual(exp_res, res)

    def test_hget(self):
        cases = (
            ('hname1', 'hval1'),
            ('hname2', 'hval2'),
            ('hname3', 'hval3'),
            ('hname4', 'hval4'),
        )

        for hname, hkey in cases:
            res = self.cli.hget(hname, hkey)
            time.sleep(0.1)

            exp_path = '{ver}/HGET/{hn}/{hk}'.format(ver=self.cli.ver, hn=hname, hk=hkey)
            self.assertEqual(exp_path, TestRedisProxyClient.request['req-path'])

            exp_qs = 'n=3&w=2&r=2'
            self.assertIn(exp_qs, TestRedisProxyClient.request['req-qs'])

            exp_res = {
                'foo': 1,
                'bar': 2,
            }

            self.assertDictEqual(exp_res, res)

    def test_set(self):
        cases = (
            ('key1', 'val1', None),
            ('key2', 'val2', 1),
            ('key3', 'val3', 2),
            ('key4', 'val4', 3),

            ('key5', 11, 4),
            ('key6', 22, 5),
        )

        for key, val, expire in cases:
            self.cli.set(key, val, expire)
            time.sleep(0.1)

            exp_path = '{ver}/SET/{k}'.format(ver=self.cli.ver, k=key)
            self.assertEqual(exp_path, TestRedisProxyClient.request['req-path'])

            exp_qs = 'n=3&w=2&r=2'
            if expire is not None:
                exp_qs += '&expire={e}'.format(e=expire)

            self.assertIn(exp_qs, TestRedisProxyClient.request['req-qs'])
            self.assertEqual(utfjson.dump(val), TestRedisProxyClient.request['req-body'])

    def test_hset(self):
        cases = (
            ('hname1', 'key1', 'val1', None),
            ('hname2', 'key2', 'val2', 1),
            ('hname3', 'key3', 'val3', 2),
            ('hname4', 'key4', 'val4', 3),

            ('hname5', 'key5', 11, 4),
            ('hname6', 'key6', 22, 5),
        )

        for hname, key, val, expire in cases:
            self.cli.hset(hname, key, val, expire=expire)
            time.sleep(0.1)

            exp_path = '{ver}/HSET/{hn}/{hk}'.format(ver=self.cli.ver, hn=hname, hk=key)
            self.assertEqual(exp_path, TestRedisProxyClient.request['req-path'])

            exp_qs = 'n=3&w=2&r=2'
            if expire is not None:
                exp_qs += '&expire={e}'.format(e=expire)

            self.assertIn(exp_qs, TestRedisProxyClient.request['req-qs'])
            self.assertEqual(utfjson.dump(val), TestRedisProxyClient.request['req-body'])

    def test_retry(self):
        # close http server
        self.tearDown()

        cases = (
            (None, 4),
            (0, 4),
            (1, 8),
            (2, 12),
            (3, 16),
            (4, 20),
        )

        sess = {'run_times': 0}

        def _mock_for_retry(req):
            sess['run_times'] += 1
            return req

        for retry_cnt, exp_cnt in cases:
            sess['run_times'] = 0
            with mock.patch('pykit.redisutil.RedisProxyClient._sign_req', side_effect=_mock_for_retry):
                try:
                    self.cli.get('foo', retry=retry_cnt)
                except:
                    pass

                try:
                    self.cli.hget('foo', 'bar', retry=retry_cnt)
                except:
                    pass

                try:
                    self.cli.set('foo', 'val', retry=retry_cnt)
                except:
                    pass

                try:
                    self.cli.hset('foo', 'bar', 'val', retry=retry_cnt)
                except:
                    pass

            self.assertEqual(exp_cnt, sess['run_times'])


class HttpHandle(BaseHTTPRequestHandler):

    def do_PUT(self):
        length = int(self.headers.getheader('Content-Length'))
        rst = []

        while length > 0:
            buf = self.rfile.read(length)
            rst.append(buf)
            length -= len(buf)

        path_res = urlparse.urlparse(self.path)

        TestRedisProxyClient.request['req-body'] = ''.join(rst)
        TestRedisProxyClient.request['req-path'] = path_res.path
        TestRedisProxyClient.request['req-qs'] = path_res.query

        body = TestRedisProxyClient.response.get('body', '')
        self.send_response(TestRedisProxyClient.response['http-status'])
        self.send_header('Content-Length', len(body))
        self.end_headers()

        if len(body) > 0:
            self.wfile.write(body)

    def do_GET(self):
        response = utfjson.dump({
            'foo': 1,
            'bar': 2,
        })

        path_res = urlparse.urlparse(self.path)

        TestRedisProxyClient.request['req-path'] = path_res.path
        TestRedisProxyClient.request['req-qs'] = path_res.query

        self.send_response(TestRedisProxyClient.response['http-status'])
        self.send_header('Content-Length', len(response))
        self.end_headers()

        self.wfile.write(response)


def child_exit():

    # to suppress error message caused by sys.exit()
    os.close(0)
    os.close(1)
    os.close(2)

    sys.exit(0)



import threading
import unittest

from kazoo.exceptions import BadVersionError

from pykit import rangeset
from pykit import threadutil
from pykit import ututil
from pykit import zktx
from pykit.zktx import COMMITTED
from pykit.zktx import STATUS

dd = ututil.dd

zk_tag = 'daocloud.io/zookeeper:3.4.10'
zk_name = 'zk_test'


class PseudoKVAccessor(dict):

    def __init__(self, *args, **kwargs):

        super(PseudoKVAccessor, self).__init__(*args, **kwargs)

        self.lock = threading.RLock()

    def get(self, key):

        with self.lock:
            return super(PseudoKVAccessor, self).get(key)

    def set(self, key, value, version):

        with self.lock:

            prev = self[key]
            if prev[1] != version:
                raise BadVersionError()

            self[key] = (value, version+1)

    def set_or_create(self, key, value, version):

        with self.lock:

            if key in self:
                return self.set(key, value, version)
            else:
                if version == -1:
                    self[key] = (value, 0)
                else:
                    raise BadVersionError()


class TxidsetAccessor(object):
    def __init__(self):
        self.lock = threading.RLock()
        self.ver = 0
        self.d = {}

    def get(self):
        with self.lock:
            for k in self.d:
                self.d[k] = rangeset.RangeSet(self.d[k])

            return self.d, self.ver

    def set(self, value, version=None):
        with self.lock:
            if version is not None and version != self.ver:
                raise BadVersionError()

            self.d = value
            self.ver = version + 1

    def set_or_create(self, key, value, version):
        return self.set(value, version)


class PseudoStorage(zktx.StorageHelper):

    conflicterror = BadVersionError

    def __init__(self):
        self.record = PseudoKVAccessor({
            'foo': ([[1, 1], [17, 17]], 0),
            'bar': ([[-1, None]], 1),
        })

        self.txidset = TxidsetAccessor()


class TestTXStorageHelper(unittest.TestCase):

    def setUp(self):
        self.sto = PseudoStorage()

    def test_get_latest(self):

        rst = self.sto.record.get('foo')
        self.assertEqual(([[1, 1], [17, 17]], 0), rst)

        rst = self.sto.record.get('bar')
        self.assertEqual(([[-1,  None]], 1), rst)

    def test_apply_record(self):

        cases = (
                (0,  'foo', 'fooval', (None,     0), '<=max txid'),
                (1,  'foo', 'fooval', (1,        0), '<=max txid'),
                (16, 'foo', 'fooval', (None,     0), '<=max txid'),
                (17, 'foo', 'fooval', (17,       0), '<=max txid'),
                (18, 'foo', 'fooval', ('fooval', 1), '>max txid'),
                (1,  'bar', 'barval', ('barval', 2), '>max txid'),
                (1,  'bar', 'barval', ('barval', 2), '<=max txid'),
        )

        for txid, key, value, expected, mes in cases:

            dd(txid, key, value, expected, mes)

            self.sto.apply_record(txid, key, value)

            rec, ver = self.sto.record.get(key)
            val_of_txid = dict(rec).get(txid)
            self.assertEqual(expected, (val_of_txid, ver))

    def test_apply_record_max_history(self):

        s, e = 100, 150
        for txid in range(s, e):
            self.sto.apply_record(txid, 'foo', txid)

        rec, ver = self.sto.record.get('foo')
        self.assertEqual(e-s, ver)
        self.assertEqual(self.sto.max_value_history, len(rec))

    def test_apply_record_concurrently(self):

        n_thread = 10
        n_repeat = 100

        self.sto.max_value_history = n_thread * n_repeat + 1

        l = threading.RLock()
        sess = {'txid': 0}
        success_tx = {-1: True}

        def _apply(ithread):
            for ii in range(n_repeat):

                with l:
                    sess['txid'] += 1
                    txid = sess['txid']

                if self.sto.apply_record(txid, 'bar', txid):
                    with l:
                        success_tx[txid] = True

        ths = [threadutil.start_daemon(_apply, args=(ii, ))
               for ii in range(n_thread)]

        for th in ths:
            th.join()

        rec, ver = self.sto.record.get('bar')
        dd(len(rec))
        dd(ver)

        self.assertEqual(set(success_tx.keys()),
                         set([x[0] for x in rec]))

        rec, ver = self.sto.record.get('bar')
        dd(rec, ver)

        self.assertLessEqual(ver, 1+n_thread*n_repeat)
        self.assertEqual(n_thread*n_repeat, rec[-1][0])
        self.assertEqual(n_thread*n_repeat, rec[-1][1])

    def _rand_txids(self):
        for ii in range(100):
            txid = int(ii % 53 * 1.3)
            yield txid

    def test_add_to_txidset(self):

        self.assertRaises(KeyError, self.sto.add_to_txidset, 'foo', 1)

        expected = {txid: True
                    for txid in self._rand_txids()}

        for status in STATUS:
            for txid in self._rand_txids():
                self.sto.add_to_txidset(status, txid)

            dd(sorted(expected.keys()))

            rst, ver = self.sto.txidset.get()
            dd(rst[status])

            for txid in expected:
                self.assertTrue(rst[status].has(txid))

            for txid in range(100):
                if txid not in expected:
                    self.assertFalse(rst[status].has(txid))

    def test_add_to_txidset_concurrent(self):

        expected = {txid: True
                    for txid in self._rand_txids()}

        n_thread = 2
        status = COMMITTED

        def _add():
            for txid in self._rand_txids():
                self.sto.add_to_txidset(status, txid)

        for th in [threadutil.start_daemon(_add)
                   for _ in range(n_thread)]:

            th.join()

        rst, ver = self.sto.txidset.get()
        dd(rst[status])

        for txid in expected:
            self.assertTrue(rst[status].has(txid))

        for txid in range(100):
            if txid not in expected:
                self.assertFalse(rst[status].has(txid))



# from .mysqlconnpool import (MysqlConnectionPool,)
from .mysqlconnpool import (
    make,
    conn_query,
    MysqlConnectionPool,
)

__all__ = [
    'make',
    'conn_query',
    'MysqlConnectionPool',
]



#!/usr/bin/env python
# coding: utf-8

import time

from pykit import rangeset
from pykit import threadutil
from pykit import utfjson
from pykit import ututil
from pykit import zktx
from pykit.zktx import ABORTED
from pykit.zktx import COMMITTED
from pykit.zktx import PURGED
from pykit.zktx import ConnectionLoss
from pykit.zktx import Deadlock
from pykit.zktx import HigherTXApplied
from pykit.zktx import NotLocked
from pykit.zktx import TXError
from pykit.zktx import TXTimeout
from pykit.zktx import UnlockNotAllowed
from pykit.zktx import UserAborted
from pykit.zktx import ZKTransaction
from pykit.zktx.test import base

dd = ututil.dd

zkhost = '127.0.0.1:2181'


class TXBase(base.ZKTestBase):

    def setUp(self):
        super(TXBase, self).setUp()

        self.zk.create('tx', 'tx')
        self.zk.create('lock', '{}')
        self.zk.create('tx/txidset', '{}')
        self.zk.create('tx/alive', '{}')
        self.zk.create('tx/txid_maker', '{}')
        self.zk.create('tx/journal', '{}')
        self.zk.create('tx/state', '{}')
        self.zk.create('record', '{}')


class TestTX(TXBase):

    def test_single_record(self):

        tx = ZKTransaction(zkhost)
        tx.begin()

        foo = tx.lock_get('foo')
        self.assertEqual('foo', foo.k)
        self.assertIsNone(foo.v)
        foo.v = 1

        tx.set(foo)

        tx.commit()

        rst, ver = self.zk.get('record/foo')
        self.assertEqual([[-1, None], [1, 1]], utfjson.load(rst))

        rst, ver = self.zk.get('tx/txidset')
        self.assertEqual({COMMITTED: [[1, 2]],
                          ABORTED: [],
                          PURGED: [],
                          }, utfjson.load(rst))

    def test_deref(self):

        with ZKTransaction(zkhost) as t1:

            foo = t1.lock_get('foo')
            foo.v = {'foo': 'foo'}
            t1.set(foo)
            t1.commit()

        with ZKTransaction(zkhost) as t1:

            # When retrieving a dict, it should be deep-copied.
            # All zktx always though modified == original
            foo = t1.lock_get('foo')
            foo.v['foo'] = 'bar'
            t1.set(foo)
            t1.commit()

        rst, ver = self.zk.get('record/foo')
        self.assertEqual([[-1, None], [1, {'foo': 'foo'}], [2, {'foo': 'bar'}]],
                         utfjson.load(rst))

    def test_with_statement(self):

        with ZKTransaction(zkhost) as t1:

            foo = t1.lock_get('foo')
            foo.v = 1

            t1.set(foo)

            t1.commit()

        rst, ver = self.zk.get('record/foo')
        self.assertEqual([[-1, None], [1, 1]], utfjson.load(rst))

        rst, ver = self.zk.get('tx/txidset')
        self.assertEqual({COMMITTED: [[1, 2]],
                          ABORTED: [],
                          PURGED: [],
                          }, utfjson.load(rst))

    def test_lock_get_twice(self):

        with ZKTransaction(zkhost) as t1:

            foo = t1.lock_get('foo')
            foo.v = 1
            t1.set(foo)

            f2 = t1.lock_get('foo')
            self.assertEqual(1, f2.v)
            f2.v = 2
            t1.set(f2)

            t1.commit()

        rst, ver = self.zk.get('record/foo')
        self.assertEqual([[-1, None], [1, 2]], utfjson.load(rst))

    def test_lock_get_latest(self):

        with ZKTransaction(zkhost) as t1:

            foo = t1.lock_get('foo')
            self.assertEqual(None, foo.v)

            foo.v = 1

            # lock_get retrieves original value without set
            f1 = t1.lock_get('foo')
            self.assertEqual(None, f1.v)

            t1.set(foo)

            # lock_get can retrieve latest value after set
            f2 = t1.lock_get('foo')
            self.assertEqual(1, f2.v)

            t1.commit()

        rst, ver = self.zk.get('record/foo')
        self.assertEqual([[-1, None], [1, 1]], utfjson.load(rst))

    def test_lock_get_not_latest(self):

        with ZKTransaction(zkhost) as t1:

            foo = t1.lock_get('foo')
            foo.v = 1
            t1.set(foo)

            # lock_get retrieves original value when latest is False
            f2 = t1.lock_get('foo', latest=False)
            self.assertEqual(None, f2.v)

            t1.commit()

        rst, ver = self.zk.get('record/foo')
        self.assertEqual([[-1, None], [1, 1]], utfjson.load(rst))

    def test_noblocking_lock_get(self):

        with ZKTransaction(zkhost) as t1:

            t1.lock_get('foo')
            f2 = t1.lock_get('foo', blocking=False)
            self.assertIsNotNone(f2)

            with ZKTransaction(zkhost) as t2:

                f4 = t2.lock_get('foo', blocking=False)
                self.assertIsNone(f4)

    def test_unlock(self):

        with ZKTransaction(zkhost) as t1:

            foo = t1.lock_get('foo')
            t1.unlock(foo)

            self.assertEqual({}, t1.got_keys)

            # re-get is ok
            foo = t1.lock_get('foo')
            t1.unlock(foo)

            with ZKTransaction(zkhost) as t2:
                # t2 can lock a unlocked key
                t2.lock_get('foo')

    def test_unlock_nonlocked(self):

        with ZKTransaction(zkhost) as t1:

            foo = t1.lock_get('foo')
            t1.unlock(foo)
            self.assertRaises(NotLocked, t1.unlock, foo)

    def test_unlock_changed_record(self):

        with ZKTransaction(zkhost) as t1:

            foo = t1.lock_get('foo')
            t1.set(foo)
            self.assertRaises(UnlockNotAllowed, t1.unlock, foo)

    def test_run_tx(self):

        def _tx(tx):
            foo = tx.lock_get('foo')
            foo.v = 100
            tx.set(foo)
            tx.commit()

        status, txid = zktx.run_tx(zkhost, _tx)
        self.assertEqual((COMMITTED, 1), (status, txid))

        t = ZKTransaction(zkhost)
        rst, ver = t.zkstorage.record.get('foo')
        self.assertEqual(100, rst[-1][1])

        rst, ver = self.zk.get('tx/txidset')
        self.assertEqual({COMMITTED: [[1, 2]],
                          ABORTED: [],
                          PURGED: [],
                          }, utfjson.load(rst))

    def test_run_tx_timeout(self):

        def _tx(tx):
            tx.lock_get('foo')
            time.sleep(0.5)
            tx.commit()

        try:
            zktx.run_tx(zkhost, _tx, timeout=0.4)
            self.fail('TXTimeout expected')
        except TXTimeout as e:
            dd(repr(e))

    def test_run_tx_conn_loss(self):

        def _tx(tx):
            tx.lock_get('foo')
            tx.zke.stop()
            tx.commit()

        try:
            zktx.run_tx(zkhost, _tx)
            self.fail('ConnectionLoss expected')
        except ConnectionLoss as e:
            dd(repr(e))

    def test_run_tx_retriable_error(self):

        errs = [HigherTXApplied, Deadlock, None]

        def _tx(tx):

            foo = tx.lock_get('foo')
            foo.v = 100
            tx.set(foo)

            err = errs.pop(0)
            if err is not None:
                raise err()
            else:
                tx.commit()

        status, txid = zktx.run_tx(zkhost, _tx)
        # 2 error tried and one commit
        self.assertEqual((COMMITTED, 3), (status, txid))

        t = ZKTransaction(zkhost)
        rst, ver = t.zkstorage.record.get('foo')
        self.assertEqual(100, rst[-1][1])

    def test_run_tx_unretriable_error(self):

        sess = dict(n=0)
        errs = [UserAborted, TXTimeout, ConnectionLoss, None]

        def _tx(tx, err):

            sess['n'] += 1

            foo = tx.lock_get('foo')
            foo.v = 100
            tx.set(foo)

            if err is not None:
                raise err()
            else:
                tx.commit()

        for err in errs:
            try:
                zktx.run_tx(zkhost, _tx, args=(err, ))
            except TXError as e:
                dd(repr(e))

        self.assertEqual(len(errs), sess['n'])

        tx = ZKTransaction(zkhost)
        rst, ver = tx.zkstorage.record.get('foo')
        self.assertEqual(100, rst[-1][1])

    def test_sequential(self):

        n_tx = 10
        k = 'foo'

        for ii in range(n_tx):

            with ZKTransaction(zkhost) as t1:

                foo = t1.lock_get(k)
                foo.v = foo.v or 0
                foo.v += 1

                t1.set(foo)

                t1.commit()

        t = ZKTransaction(zkhost)

        rst, ver = t.zkstorage.record.get(k)
        dd(rst)
        self.assertEqual(n_tx, rst[-1][1])

        rst, ver = t.zkstorage.txidset.get()
        dd(rst)
        self.assertEqual(n_tx, rst[COMMITTED].length())

    def test_concurrent_single_record(self):

        n_tx = 10

        def _tx():
            while True:
                try:
                    with ZKTransaction(zkhost) as t1:

                        foo = t1.lock_get('foo')
                        foo.v = foo.v or 0
                        foo.v += 1

                        t1.set(foo)
                        t1.commit()
                        return

                except (Deadlock,
                        HigherTXApplied) as e:
                    dd(repr(e))
                    continue

        for th in [threadutil.start_daemon(_tx)
                   for i in range(n_tx)]:

            th.join()

        t = ZKTransaction(zkhost)

        rst, ver = t.zkstorage.record.get('foo')
        dd(rst)
        self.assertEqual(n_tx, rst[-1][1])

        rst, ver = t.zkstorage.txidset.get()
        dd(rst)
        self.assertEqual(n_tx, rst[COMMITTED].length())

    def test_concurrent_3_record(self):

        n_tx = 10
        ks = ('foo', 'bar', 'duu')

        def _tx(i):
            while True:
                try:
                    with ZKTransaction(zkhost) as t1:

                        for ii in range(len(ks)):
                            k = ks[(ii+i) % len(ks)]

                            foo = t1.lock_get(k)
                            foo.v = foo.v or 0
                            foo.v += 1

                            t1.set(foo)

                        t1.commit()
                        dd(str(t1) + ' committed')
                        return

                except (Deadlock,
                        HigherTXApplied) as e:
                    dd(str(t1) + ': ' + repr(e))
                    continue
                except TXTimeout as e:
                    dd(str(t1) + ': ' + repr(e))
                    raise

        with ututil.Timer() as tt:

            for th in [threadutil.start_daemon(_tx, args=(i, ))
                       for i in range(n_tx)]:

                th.join()

            dd('3 key 10 thread: ', tt.spent())

        t = ZKTransaction(zkhost)

        for k in ks:
            rst, ver = t.zkstorage.record.get(k)
            dd(rst)
            self.assertEqual(n_tx, rst[-1][1])

        rst, ver = t.zkstorage.txidset.get()
        dd(rst)
        self.assertEqual(n_tx, rst[COMMITTED].length())

        # check aborted txidset
        sm = rangeset.union(rst[COMMITTED], rst[ABORTED])
        self.assertEqual(rst[COMMITTED][-1][-1] - 1, sm.length())

    def test_redo_dead_tx_without_journal(self):

        t1 = ZKTransaction(zkhost)
        t1.begin()
        t1.lock_get('foo')

        t1.zke.stop()

        with ZKTransaction(zkhost) as t2:

            foo = t2.lock_get('foo')
            foo.v = foo.v or 0
            foo.v += 1

            t2.set(foo)
            t2.commit()

        t = ZKTransaction(zkhost)

        rst, ver = t.zkstorage.record.get('foo')
        dd(rst)
        self.assertEqual(1, rst[-1][1])

        rst, ver = t.zkstorage.txidset.get()
        dd(rst)
        self.assertEqual([1, 2], rst[ABORTED][0])
        self.assertEqual([2, 3], rst[COMMITTED][0])

    def test_abort(self):

        with ZKTransaction(zkhost) as t1:
            foo = t1.lock_get('foo')
            foo.v = 100
            t1.set(foo)
            t1.abort()

        t = ZKTransaction(zkhost)

        rst, ver = t.zkstorage.record.get('foo')
        dd(rst)
        self.assertEqual(None, rst[-1][1])

        rst, ver = t.zkstorage.txidset.get()
        dd(rst)
        self.assertEqual([1, 2], rst[ABORTED][0])


class TestTXState(TXBase):

    def test_committed(self):

        t = ZKTransaction(zkhost)
        txid = None

        with ZKTransaction(zkhost) as t1:

            txid = t1.txid

            # not set yet
            self.assertIsNone(t.zkstorage.state.get(txid)[0])

            foo = t1.lock_get('foo')
            foo.v = 100
            t1.set(foo)
            t1.set_state('bar')

            val, ver = t.zkstorage.state.get(txid)
            self.assertEqual({'got_keys': ['foo'], 'data': 'bar'}, val)

            t1.commit()

        self.assertIsNone(t.zkstorage.state.get(txid)[0])

    def test_user_abort(self):

        t = ZKTransaction(zkhost)
        txid = None

        with ZKTransaction(zkhost) as t1:

            txid = t1.txid
            t1.set_state('bar')

            val, ver = t.zkstorage.state.get(txid)
            self.assertEqual({'got_keys': [], 'data': 'bar'}, val)

            t1.abort()

        self.assertIsNone(t.zkstorage.state.get(txid)[0])

    def test_exception(self):

        # tx raised by other exception is recoverable

        t = ZKTransaction(zkhost)
        txid = None

        try:
            with ZKTransaction(zkhost) as t1:

                txid = t1.txid
                t1.set_state('bar')

                val, ver = t.zkstorage.state.get(txid)
                self.assertEqual({'got_keys': [], 'data': 'bar'}, val)

                raise ValueError('foo')

        except ValueError:
            pass

        val, ver = t.zkstorage.state.get(txid)
        self.assertEqual({'got_keys': [], 'data': 'bar'}, val)

    def test_timeout(self):

        # timeout on waiting is recoverable

        t = ZKTransaction(zkhost)
        txid = None

        try:
            with ZKTransaction(zkhost, timeout=0.5) as t1:

                txid = t1.txid

                # t.txid is higher
                t.begin()
                t.lock_get('foo')

                t1.set_state('bar')
                t1.lock_get('foo')  # should timeout waiting for higher txid

        except TXTimeout:
            pass

        val, ver = t.zkstorage.state.get(txid)
        self.assertEqual({'got_keys': [], 'data': 'bar'}, val)

    def test_deadlock(self):

        # deadlock is not recoverable

        t = ZKTransaction(zkhost)
        # t.txid is lower
        t.begin()
        t.lock_get('foo')

        def _commit():
            t.commit()

        txid = None

        try:
            with ZKTransaction(zkhost, timeout=0.5) as t1:

                txid = t1.txid

                # lock another key first to produce deadlock
                t1.lock_get('woo')

                t1.set_state('bar')
                threadutil.start_daemon(_commit, after=0.2)
                t1.lock_get('foo')  # should deadlock waiting for higher txid

        except Deadlock:
            pass

        t = ZKTransaction(zkhost)
        self.assertIsNone(t.zkstorage.state.get(txid)[0])

    def test_uncommitted(self):

        # uncommitted is recoverable

        t = ZKTransaction(zkhost)
        txid = None

        with ZKTransaction(zkhost, timeout=0.5) as t1:

            txid = t1.txid
            t1.lock_get('foo')
            t1.set_state('bar')

        val, ver = t.zkstorage.state.get(txid)
        self.assertEqual({'got_keys': ['foo'], 'data': 'bar'}, val)

    def test_list_recoverable(self):

        txid = None
        with ZKTransaction(zkhost, timeout=0.5) as t1:

            txid = t1.txid
            t1.lock_get('foo')
            t1.set_state('bar')

            with ZKTransaction(zkhost, timeout=0.5) as t2:
                txid = t2.txid
                t2.lock_get('foo2')
                t2.set_state('bar2')

            self.assertEqual([(txid, 'bar2')],
                             [x for x in zktx.list_recoverable(zkhost)])

    def test_recover(self):

        txid = None
        with ZKTransaction(zkhost, timeout=0.5) as t1:

            txid = t1.txid
            t1.lock_get('foo')
            t1.set_state('bar')

        with ZKTransaction(zkhost, txid=txid, timeout=0.5) as t2:
            st = t2.get_state()
            self.assertEqual('bar', st)

            try:
                with ZKTransaction(zkhost, txid=txid, timeout=0.5) as t3:
                    dd(t3)
                self.fail("expected TXTimeout")
            except TXTimeout:
                pass

    def test_redo_all(self):

        # txid=1 committed
        # txid=2 nothing
        # txid=3 has journal
        # txid=4 no tx but has state
        # txid=5 committed

        # txid=1
        with ZKTransaction(zkhost) as t1:
            foo = t1.lock_get('foo')
            foo.v = 1
            t1.set(foo)
            t1.commit()

        t = ZKTransaction(zkhost)

        # make txid=2, leave it a hole
        t.zke.set(t.zke._zkconf.txid_maker(), 'x')

        # make txid=3
        t.zke.set(t.zke._zkconf.txid_maker(), 'x')
        txid = 3
        # fake a journal
        t.zkstorage.journal.create(txid, {'bar': 3})

        # txid=4, has state
        t.zke.set(t.zke._zkconf.txid_maker(), 'x')
        t.zkstorage.state.create(4, {'xx': 'yy'})

        # txid=5, committed
        with ZKTransaction(zkhost) as t5:
            foo = t5.lock_get('foo')
            foo.v = 5
            t5.set(foo)
            t5.commit()

        sets, ver = t.zkstorage.txidset.get()
        dd('init txidset:', sets)
        self.assertEqual({'PURGED': [], 'ABORTED': [], 'COMMITTED': [[1, 2], [5, 6]]}, sets)

        t.redo_all_dead_tx()  # redo txid=2, abort
        sets, ver = t.zkstorage.txidset.get()
        dd('after redo 2:', sets)
        self.assertEqual({'PURGED': [], 'ABORTED': [[2, 3]], 'COMMITTED': [[1, 2], [5, 6]]}, sets)

        t.redo_all_dead_tx()  # redo txid=3, committed
        sets, ver = t.zkstorage.txidset.get()
        dd('after redo 3:', sets)
        self.assertEqual({'PURGED': [], 'ABORTED': [[2, 3]], 'COMMITTED': [[1, 2], [3, 4], [5, 6]]}, sets)

        t.redo_all_dead_tx()  # redo txid=4, can not redo a tx with state
        sets, ver = t.zkstorage.txidset.get()
        dd('after redo 4, ignored 4:', sets)
        self.assertEqual({'PURGED': [], 'ABORTED': [[2, 3]], 'COMMITTED': [[1, 2], [3, 4], [5, 6]]}, sets)

    def test_unlock_with_state(self):

        # NotLocked is retrieable

        txid = None

        try:
            with ZKTransaction(zkhost, timeout=0.5) as t1:

                txid = t1.txid
                t1.set_state('bar')

                foo = t1.lock_get('foo')
                t1.unlock(foo)
                t1.unlock(foo)

        except NotLocked:
            pass

        t = ZKTransaction(zkhost)
        val, ver = t.zkstorage.state.get(txid)
        self.assertEqual({'got_keys': [], 'data': 'bar'}, val)

        # UnlockNotAllowed is retrieable

        txid = None

        try:
            with ZKTransaction(zkhost, timeout=0.5) as t1:

                txid = t1.txid
                t1.set_state('bar')

                foo = t1.lock_get('foo')
                t1.set(foo)
                t1.unlock(foo)

        except UnlockNotAllowed:
            pass

        t = ZKTransaction(zkhost)
        val, ver = t.zkstorage.state.get(txid)
        self.assertEqual({'got_keys': [], 'data': 'bar'}, val)



import argparse
import os
import sys

from . import portlock
from .. import proc

if __name__ == "__main__":

    parser = argparse.ArgumentParser(description='portlock tool')

    parser.add_argument('--lock',    '-l', type=str,   action='store', help='lock name')
    parser.add_argument('--timeout', '-t', type=float, action='store', help='lock wait timeout, in second')
    parser.add_argument('--script',  '-c', type=str,   action='store', help='execute a script in string')
    parser.add_argument('command',         type=str, nargs='*',        help='shell command to execute')

    args = parser.parse_args()

    timeout = args.timeout
    if timeout is None:
        timeout = 1

    try:
        with portlock.Portlock(args.lock, timeout=timeout):

            if args.script is not None:
                rc, out, err = proc.shell_script(args.script)
            else:
                rc, out, err = proc.command(*args.command)

            os.write(1, out)
            os.write(2, err)
            sys.exit(rc)

    except portlock.PortlockTimeout as e:
        os.write(2, 'portlock {lock} timeout: {timeout} sec'.format(
                lock=args.lock,
                timeout=timeout))
        sys.exit(1)



#!/usr/bin/env python2
# coding: utf-8

import os
import re
import time
import unittest

from pykit import logutil
from pykit import threadutil
from pykit import timeutil
from pykit import ututil
from pykit.logcollector import collector

dd = ututil.dd

this_base = os.path.dirname(__file__)

standard_level = {
    'ERROR': 'error',
    'WARNING': 'warn',
    'INFO': 'info',
}


def is_first_line(line):
    return True


def parse(log_str):
    r = re.match('^\[(.+?),(.+?),(.+?),(.+?),(\d+?),(\w+?)]', log_str)

    time_str = r.group(1)

    log_dt = timeutil.parse(time_str, 'mysql')
    log_ts = int(time.mktime(log_dt.timetuple()))

    source_file = r.group(4)
    line_number = int(r.group(5))
    level = standard_level[r.group(6)]

    log_info = {
        'log_ts': log_ts,
        'level': level,
        'source_file': source_file,
        'line_number': line_number,
    }

    return log_info


class TestLogcollector(unittest.TestCase):

    def _clean(self):
        try:
            os.unlink(os.path.join(this_base, 'test_log.out'))
        except Exception as e:
            dd(repr(e))

    def setUp(self):
        self._clean()

    def tearDown(self):
        self._clean()

    def test_basic(self):
        logger = logutil.make_logger(base_dir=this_base,
                                     log_name='test_log')

        def log():
            start_time = time.time()

            while True:
                logger.info('info')
                logger.warn('warn')
                logger.error('error')

                if time.time() > start_time + 2.5:
                    break

                time.sleep(0.01)

        log_th = threadutil.start_daemon_thread(log)

        log_entries = []

        def send_log(log_entry):
            log_entries.append(log_entry)

        kwargs = {
            'node_id': '123abc',
            'node_ip': '1.2.3.4',
            'send_log': send_log,
            'conf': {
                'my_test_log': {
                    'file_path': os.path.join(this_base, 'test_log.out'),
                    'level': ['error'],
                    'is_first_line': is_first_line,
                    'parse': parse,
                },
            },
        }

        threadutil.start_daemon_thread(collector.run, kwargs=kwargs)

        log_th.join()

        time.sleep(2)

        self.assertEqual(3, len(log_entries))

        dd(log_entries[0]['count'])
        dd(log_entries[1]['count'])
        dd(log_entries[2]['count'])
        self.assertAlmostEqual(100, log_entries[1]['count'], delta=15)

        self.assertEqual('error', log_entries[0]['level'])
        self.assertEqual('my_test_log', log_entries[0]['log_name'])
        self.assertEqual('test_log.out', log_entries[0]['log_file'])



from .logutil import (
    add_std_handler,
    deprecate,
    get_datefmt,
    get_fmt,
    get_root_log_fn,
    make_file_handler,
    make_formatter,
    make_logger,
    set_logger_level,
    stack_format,
    stack_list,
    stack_str,
)

__all__ = [
    'add_std_handler',  # used
    'deprecate',
    'get_datefmt',
    'get_fmt',  # used
    'get_root_log_fn',  # used
    'make_file_handler',
    'make_formatter',
    'make_logger',
    'set_logger_level',
    'stack_format',
    'stack_list',
    'stack_str',  # used
]



from pykit.p3json.test import PyTest


# from http://json.org/JSON_checker/test/pass1.json
JSON = r'''
[
    "JSON Test Pattern pass1",
    {"object with 1 member":["array with 1 element"]},
    {},
    [],
    -42,
    true,
    false,
    null,
    {
        "integer": 1234567890,
        "real": -9876.543210,
        "e": 0.123456789e-12,
        "E": 1.234567890E+34,
        "":  23456789012E66,
        "zero": 0,
        "one": 1,
        "space": " ",
        "quote": "\"",
        "backslash": "\\",
        "controls": "\b\f\n\r\t",
        "slash": "/ & \/",
        "alpha": "abcdefghijklmnopqrstuvwyz",
        "ALPHA": "ABCDEFGHIJKLMNOPQRSTUVWYZ",
        "digit": "0123456789",
        "0123456789": "digit",
        "special": "`1~!@#$%^&*()_+-={':[,]}|;.</>?",
        "hex": "\u0123\u4567\u89AB\uCDEF\uabcd\uef4A",
        "true": true,
        "false": false,
        "null": null,
        "array":[  ],
        "object":{  },
        "address": "50 St. James Street",
        "url": "http://www.JSON.org/",
        "comment": "// /* <!-- --",
        "# -- --> */": " ",
        " s p a c e d " :[1,2 , 3

,

4 , 5        ,          6           ,7        ],"compact":[1,2,3,4,5,6,7],
        "jsontext": "{\"object with 1 member\":[\"array with 1 element\"]}",
        "quotes": "&#34; \u0022 %22 0x22 034 &#x22;",
        "\/\\\"\uCAFE\uBABE\uAB98\uFCDE\ubcda\uef4A\b\f\n\r\t`1~!@#$%^&*()_+-=[]{}|;:',./<>?"
: "A key can be any string"
    },
    0.5 ,98.6
,
99.44
,

1066,
1e1,
0.1e1,
1e-1,
1e00,2e+00,2e-00
,"rosebud"]
'''

class TestPass1(object):
    def test_parse(self):
        # test in/out equivalence and parsing
        res = self.loads(JSON)
        # XXX for python2 json compatibility: use ensure_ascii=False to dumps encoded string.
        out = self.dumps(res, ensure_ascii=False)
        self.assertEqual(res, self.loads(out))


class TestPyPass1(TestPass1, PyTest): pass
